from __future__ import annotations

import abc
from collections import namedtuple
from functools import lru_cache

import e3.log
import stevedore


class PlatformDBPlugin(metaclass=abc.ABCMeta):
    """Plugin API to extend the platform knowledge base.

    To create a plugin, override this class and the method ``update_db``. In
    ``update_db`` modify the values of self.cpu_info, self.os_info,
    self.platform_info, self.build_targets, and self.host_guess.

    Then add an entry points in your package in the group e3.platform_db and
    reference your new class. e.g.::

        entry_points={
            'e3.platform_db': [
                'my_db = e3.mypackage.platform_db:MyPlatformDBPlugin']}
    """

    def __init__(self, cpu_info, os_info, platform_info, build_targets, host_guess):
        self.cpu_info = cpu_info
        self.os_info = os_info
        self.platform_info = platform_info
        self.build_targets = build_targets
        self.host_guess = host_guess

    @abc.abstractmethod
    def update_db(self):
        pass  # all: no cover


class AmberCPUSupport(PlatformDBPlugin):
    """Plugin example adding support for Amber CPUs."""

    def update_db(self):
        """Add support for Amber CPUs."""
        self.cpu_info.update(
            {
                "amber23": {"endian": "little", "bits": 32},
                "amber25": {"endian": "little", "bits": 32},
            }
        )


@lru_cache()
def get_knowledge_base():
    """Load the knowledge base, including all content from plugins.

    :return: The knowledge base with the keys (cpu_info, os_info,
        platform_info, build_targets, host_guess)
    """
    from e3.platform_db.knowledge_base import (
        CPU_INFO,
        OS_INFO,
        PLATFORM_INFO,
        BUILD_TARGETS,
        HOST_GUESS,
    )

    e3.log.debug("loading knownledge base")
    # Load all platform_db plugins
    ext = stevedore.ExtensionManager(
        namespace="e3.platform_db",
        invoke_on_load=True,
        invoke_args=(CPU_INFO, OS_INFO, PLATFORM_INFO, BUILD_TARGETS, HOST_GUESS),
    )

    plugin_names = ext.names()

    if plugin_names:
        # Some plugins have been found
        e3.log.debug("loading knowledge base plugins %s", ",".join(plugin_names))

        ext.map_method("update_db")
    return namedtuple(
        "KnowledgeBase",
        ("cpu_info", "os_info", "platform_info", "build_targets", "host_guess"),
    )(CPU_INFO, OS_INFO, PLATFORM_INFO, BUILD_TARGETS, HOST_GUESS)



from __future__ import annotations

import inspect
import os
import sys
import types

import yaml
from typing import TYPE_CHECKING

import e3.hash
import e3.log
from e3.anod.error import SandBoxError
from e3.anod.spec import __version__
from e3.fs import ls

logger = e3.log.getLogger("anod.loader")

if TYPE_CHECKING:
    from typing import Callable, Dict, List, Optional
    from types import ModuleType
    from e3.anod.spec import Anod


class AnodSpecRepository:
    """Anod spec repository.

    :ivar specs: A dictionary of AnodModule objects, indexed by spec name
        (without the spec filename's extension).
    :vartype specs: dict[e3.anod.loader.AnodModule]

    The object represent a set of anod specifications along with their data
    files.
    """

    def __init__(self, spec_dir: str, spec_config: Optional[dict] = None):
        """Initialize an AnodSpecRepository.

        :param spec_dir: directory containing the anod specs.
        :param spec_config: dictionary containing the configuration for this
            AnodSpecRepository
        """
        logger.debug("initialize spec repository (%s)", spec_dir)

        if not os.path.isdir(spec_dir):
            raise SandBoxError("spec directory %s does not exist" % spec_dir)
        self.spec_dir = spec_dir
        self.api_version = __version__
        self.specs = {}
        self.repos: Dict[str, Dict[str, str]] = {}

        # Look for all spec files and data files
        spec_list = {
            os.path.basename(os.path.splitext(k)[0]): {"path": k, "data": []}
            for k in ls(os.path.join(self.spec_dir, "*.anod"), emit_log_record=False)
        }
        logger.debug("found %s specs", len(spec_list))

        # API == 1.4
        yaml_files = ls(os.path.join(self.spec_dir, "*.yaml"), emit_log_record=False)
        data_list = [os.path.basename(k)[:-5] for k in yaml_files]
        logger.debug("found %s yaml files API 1.4 compatible", len(data_list))

        # Match yaml files with associated specifications
        for data in data_list:
            candidate_specs = [
                spec_file for spec_file in spec_list if data.startswith(spec_file)
            ]
            # We pick the longuest spec name
            candidate_specs.sort(key=len)
            if candidate_specs:
                spec_list[candidate_specs[-1]]["data"].append(data)  # type: ignore

        # Find yaml files that are API >= 1.5 compatible
        new_yaml_files = ls(
            os.path.join(self.spec_dir, "*", "*.yaml"), emit_log_record=False
        )

        for yml_f in new_yaml_files:
            associated_spec = os.path.basename(os.path.dirname(yml_f))

            # Keep only the yaml files associated with an .anod file
            if associated_spec in spec_list:
                # We're recording the relative path without the extension
                suffix, _ = os.path.splitext(os.path.basename(yml_f))

                spec_list[associated_spec]["data"].append(  # type: ignore
                    os.path.join(associated_spec, suffix)
                )

        # Create AnodModule objects
        for name, value in spec_list.items():
            self.specs[name] = AnodModule(name, **value)  # type: ignore

        # Load config/repositories.yaml
        repo_file = os.path.join(self.spec_dir, "config", "repositories.yaml")
        if os.path.isfile(repo_file):
            with open(repo_file) as fd:
                self.repos = yaml.safe_load(fd)

        # Declare spec prolog
        prolog_file = os.path.join(spec_dir, "prolog.py")
        self.prolog_dict = {"spec_config": spec_config, "__spec_repository": self}
        if os.path.exists(prolog_file):
            with open(prolog_file) as f:
                exec(compile(f.read(), prolog_file, "exec"), self.prolog_dict)

    def __contains__(self, item: str) -> bool:
        """Check by name if a spec is present in the repository.

        :return: True if present, False otherwise
        """
        return item in self.specs and self.specs[item].module is not None

    def load_all(self, ignore_errors: bool = False) -> None:
        """Load all the specs present in the repository.

        :param ignore_errors: if True don't stop on error.
        """
        for name in self.specs:
            try:
                self.specs[name].load(self)
            except SandBoxError:
                if not ignore_errors:
                    raise

    def load(self, name: str) -> Callable[..., Anod]:
        """Load an anod spec and return the corresponding Anod class.

        :param name: name of the spec to load
        """
        assert name in self.specs, "spec %s not found" % name
        return self.specs[name].load(self)


class AnodModule:
    def __init__(self, name: str, path: str, data: List[str]):
        """Initialize an AnodModule instance.

        :param name: module name
        :param path: path to the spec
        :param data: list of data files (yaml files) associated with the spec
        """
        self.name = name
        self.data = data
        self.path = path
        self.module: Optional[ModuleType] = None
        self.anod_class: Optional[Callable[..., Anod]] = None
        self.checksum = e3.hash.sha1(self.path)

    @property
    def is_loaded(self) -> bool:
        """Check if a spec is already loaded.

        :return: True if the spec is already loaded, False otherwise
        """
        return self.module is not None

    def load(self, repository: AnodSpecRepository) -> Callable[..., Anod]:
        """Load an anod specification and return the corresponding Anod class.

        :param repository: the anod spec repository of the spec file
        :raise SandBoxError: in case of failure
        """
        if self.is_loaded:
            if TYPE_CHECKING:
                assert self.anod_class is not None
            return self.anod_class

        logger.debug("loading anod spec: %s", self.name)

        # Create a new module
        mod_name = "anod_" + self.name
        anod_module = types.ModuleType(mod_name)

        try:
            with open(self.path) as fd:
                # Inject the prolog into the new module dict
                anod_module.__dict__.update(repository.prolog_dict)

                # Exec spec code
                code = compile(fd.read(), self.path, "exec")
                exec(code, anod_module.__dict__)
        except Exception as e:
            logger.error("exception: %s", e)
            logger.error("cannot load code of %s", self.name)
            raise SandBoxError(
                origin="load", message="invalid spec code for %s" % self.name
            ).with_traceback(sys.exc_info()[2])

        # At this stage we have loaded completely the module. Now we need to
        # look for a subclass of Anod. Use python inspection features to
        # achieve this.

        for members in inspect.getmembers(anod_module):
            _, value = members
            # Return the first subclass of Anod defined in this module
            if (
                inspect.isclass(value)
                and value.__module__ == mod_name
                and "Anod" in (k.__name__ for k in value.__mro__)
            ):
                # This class is a child of Anod so register it.
                # Note that even if we won't use directly the
                # module we need to keep a reference on it in order
                # to avoid garbage collector issues.
                value.spec_checksum = self.checksum

                # Give a name to our Anod class: the basename of the
                # anod spec file (without the .anod extension)
                value.name = self.name
                self.anod_class = value
                self.module = anod_module
                self.anod_class.data_files = self.data  # type: ignore
                self.anod_class.spec_dir = os.path.dirname(self.path)  # type: ignore
                self.anod_class.api_version = repository.api_version  # type: ignore
                return value

        logger.error("spec %s does not contains an Anod subclass", self.name)
        raise SandBoxError("cannot find Anod subclass in %s" % self.path, "load")


def spec(name: str) -> Callable[..., Anod]:
    """Load an Anod spec class.

    Note that two spec having the same name cannot be loaded in the same
    process as e3 keeps a cache of loaded spec using the spec basename as a
    key.
    :param name: name of the spec to load
    """
    spec_repository: Optional[AnodSpecRepository] = None
    for k in inspect.stack()[1:]:
        if "__spec_repository" in k[0].f_globals:
            spec_repository = k[0].f_globals["__spec_repository"]
            break

    assert spec_repository is not None
    return spec_repository.load(name)



import e3.electrolyt.entry_point as entry_point
import e3.electrolyt.host as host
import e3.electrolyt.plan as plan
import e3.env


def build_action(spec, build=None, host=None, target=None, board=None):
    return "build " + spec


def _get_new_plancontext(machine_name, ignore_disabled=True):
    server = host.Host(hostname=machine_name, platform="x86_64-linux", version="suse11")

    context = plan.PlanContext(server=server, ignore_disabled=ignore_disabled)
    context.register_action("build", build_action)
    return context


def _get_plan(data, content):
    myplan = plan.Plan(data, entry_point_cls={"ms_preset": entry_point.EntryPoint})
    with open("plan.txt", "w") as f:
        f.write("\n".join(content))
    myplan.load("plan.txt")
    return myplan


def test_simple_plan():
    """Test a very basic electrolyt plan."""
    context = _get_new_plancontext("myserver")

    myplan = _get_plan(
        data={"a": "a", "b": "b"},
        content=[
            "def myserver():\n",
            "    build(a)\n",
            '    build(b, build="x86-linux", host="x86-linux",',
            '          target="x86-windows")\n',
        ],
    )

    actions = context.execute(myplan, "myserver")
    assert len(actions) == 2
    assert actions[0].spec == "a"
    assert actions[1].build.platform == "x86-linux"
    assert actions[1].target.platform == "x86-windows"
    assert actions[1].action == "build"
    assert actions[1].spec == "b"


def test_plan_without_server_info():
    context = plan.PlanContext()
    context.register_action("build", build_action)
    myplan = _get_plan(data={"a": "a"}, content=["def myserver():\n", "    build(a)\n"])
    actions = context.execute(myplan, "myserver")
    assert actions[0].build.platform == e3.env.Env().build.platform


def test_plan_scope():
    context = _get_new_plancontext("myserver")

    myplan = _get_plan(
        data={},
        content=[
            "def myserver():\n",
            '    build("foo")\n'
            '    with defaults(build="x86_64-windows",\n'
            '                  host="x86-windows",\n'
            '                  target="x86-linux",\n'
            '                  weathers=["dev"]):\n'
            '        build("foo")\n'
            '        with defaults(target="x86_64-linux"):\n'
            '            build("foo")\n'
            '        with defaults(bar="bar"):\n'
            "            build(env.bar)\n"
            '    build("bar")\n',
        ],
    )

    actions = context.execute(myplan, "myserver")
    assert len(actions) == 5
    assert actions[0].spec == "foo"
    assert actions[0].build.platform == actions[0].target.platform
    assert actions[1].target.platform == "x86-linux"
    assert actions[1].build.platform == "x86_64-windows"
    assert actions[1].host.platform == "x86-windows"
    assert actions[1].weathers == ["dev"]
    assert actions[2].target.platform == "x86_64-linux"
    assert actions[3].spec == "bar"
    assert actions[3].target.platform == "x86-linux"
    assert actions[4].build.platform == actions[0].build.platform
    assert actions[4].build.platform == actions[4].target.platform


def test_board():
    """Test a very basic electrolyt plan."""
    context = _get_new_plancontext("myserver")

    myplan = _get_plan(
        data={"a": "a", "b": "b"},
        content=["def myserver():\n", '    build(a, board="bobby")\n'],
    )

    actions = context.execute(myplan, "myserver")
    assert len(actions) == 1
    assert actions[0].spec == "a"
    assert actions[0].target.machine == "bobby"


def test_entry_points():
    """Test a plan containing electrolyt entry points."""
    plan_content = [
        '@machine(name="machine1", description="Machine 1",',
        '         platform="x86_64-linux", version="rhES6")',
        "def machine1():",
        '    build("a")',
        "",
        '@machine(description="Machine 2",',
        '         platform="x86_64-linux", version="rhES6")',
        "def machine2():",
        '    build("b")',
        "",
        '@ms_preset(name="foo")',
        "def run_foo():" '    build("c")',
    ]

    myplan = _get_plan({}, plan_content)

    db = myplan.entry_points

    assert len(db) == 3
    assert db["foo"].kind == "ms_preset"
    assert db["foo"].name == "foo"
    assert db["foo"].fun.__name__ == "run_foo"
    assert db["machine1"].description == "Machine 1"
    assert db["machine2"].name == "machine2"

    context = _get_new_plancontext("machine2")
    actions = context.execute(myplan, "machine2")
    assert len(actions) == 1
    assert actions[0].spec == "b"

    ep_executed = [ep for ep in list(db.values()) if ep.executed]

    assert len(ep_executed) == 1
    assert ep_executed[0].name == "machine2"


def test_plan_disable_lines():
    """Check that lines in enabled=False blocks are disabled."""
    context = _get_new_plancontext("myserver")

    myplan = _get_plan(
        data={},
        content=[
            "def myserver():\n",
            '    build("foo")\n'
            "    with defaults(enabled=False):\n"
            '        build("bar1")\n'
            # It is not possible to "revert" the enabled=False status
            "        with defaults(enabled=True):\n"
            '            build("bar2")\n'
            '    build("bar3")\n',
        ],
    )

    actions = context.execute(myplan, "myserver")
    assert len(actions) == 2
    assert actions[0].spec == "foo"
    assert actions[1].spec == "bar3"

    context2 = _get_new_plancontext("myserver", ignore_disabled=False)
    actions2 = context2.execute(myplan, "myserver")
    assert len(actions2) == 4
    assert actions2[0].spec == "foo"
    assert actions2[1].spec == "bar1"
    assert actions2[2].spec == "bar2"
    assert actions2[3].spec == "bar3"


def test_plan_disable_lines_with_conditional_constants():
    """Check that lines in enabled=False blocks are disabled."""
    context = _get_new_plancontext("myserver")

    myplan = _get_plan(
        data={},
        content=[
            "def myserver():\n",
            '    build("foo")\n'
            '    enabled = cond("old", date=lambda x: x.year == 2001)\n'
            "    with defaults(enabled=enabled):\n"
            '        build("bar")\n',
        ],
    )

    actions = context.execute(myplan, "myserver")
    assert len(actions) == 1
    assert actions[0].spec == "foo"

    context2 = _get_new_plancontext("myserver", ignore_disabled=False)
    actions2 = context2.execute(myplan, "myserver")
    assert len(actions2) == 2
    assert actions2[0].spec == "foo"
    assert actions2[1].spec == "bar"

    assert not actions2[1].plan_args["enabled"]
    list(myplan.toggleable_bool_group.shuffle())
    assert actions2[1].plan_args["enabled"]



from dataclasses import dataclass
from e3.config import Config, ConfigSection


def test_config():
    with open("e3.toml", "w") as f:
        f.write(
            '[e3]\nfoo = 2\n  [e3.log]\n  pretty = false\n  stream_fmt = "%(message)s"'
        )

    Config.load_file("e3.toml")
    print(Config.data)

    @dataclass
    class MyConfig(ConfigSection):
        title = "e3.log"
        pretty: bool = True

    config = MyConfig.load()
    print(config)
    assert not config.pretty



"""High-Level file manipulation."""
from __future__ import annotations

import fnmatch
import glob
import hashlib
import itertools
import os
import shutil
import stat
import sys
from collections import namedtuple
from typing import TYPE_CHECKING

import e3
import e3.error
import e3.log
import e3.os.fs

logger = e3.log.getLogger("fs")

if TYPE_CHECKING:
    from typing import Callable, Iterable, List, Optional, Sequence, Tuple, Union


class FSError(e3.error.E3Error):
    pass


def cp(
    source: str,
    target: str,
    copy_attrs: bool = True,
    recursive: bool = False,
    preserve_symlinks: bool = False,
) -> None:
    """Copy files.

    :param source: a glob pattern
    :param target: target file or directory. If the source resolves as
        several files then target should be a directory
    :param copy_attrs: If True, also copy all the file attributes such as
        mode, timestamps, ownership, etc.
    :param recursive: If True, recursive copy. This also preserves
        attributes; if copy_attrs is False, a warning is emitted.
    :param preserve_symlinks: if True symlinks are recreated in the
        destination folder
    :raise FSError: if an error occurs
    """
    switches = ""
    if copy_attrs:
        switches += " -p"
    if recursive:
        switches += " -r"
    logger.debug("cp %s %s->%s", switches, source, target)

    if recursive and not copy_attrs:
        logger.warning("recursive copy always preserves file attributes")

    # Compute file list and number of file to copy
    file_list = ls(source, emit_log_record=False)
    file_number = len(file_list)

    if file_number == 0:
        # If there is no source files raise an error
        raise FSError(origin="cp", message='can\'t find files matching "%s"' % source)
    elif file_number > 1:
        # If we have more than one file to copy then check that target is a
        # directory
        if not os.path.isdir(target):
            raise FSError(origin="cp", message="target should be a directory")

    for f in file_list:
        try:
            if os.path.isdir(target):
                f_dest = os.path.join(target, os.path.basename(f))
            else:
                f_dest = target

            if recursive and os.path.isdir(f):
                shutil.copytree(f, f_dest, symlinks=preserve_symlinks)
            elif preserve_symlinks and os.path.islink(f):  # windows: no cover
                linkto = os.readlink(f)
                os.symlink(linkto, f_dest)
            elif copy_attrs:
                shutil.copy2(f, f_dest)
            else:
                shutil.copy(f, f_dest)
        except Exception as e:
            logger.error(e, exc_info=True)
            raise FSError(
                origin="cp", message="error occurred while copying %s" % f
            ).with_traceback(sys.exc_info()[2])


def directory_content(
    path: str, include_root_dir: bool = False, unixpath: bool = False
) -> List[str]:
    """Return the complete directory content (recusrsively).

    :param path: path for the which the content should be returned
    :param include_root_dir: if True include the root directory in the paths
        returned by the function. Otherwise return relative paths
    :param unixpath: if True return unix compatible paths (calling unixpath on
        all elements returned
    :return: a list of of path. Note that directories will end with a path
        separator
    """
    result = []
    for root, dirs, files in os.walk(path):
        for f in files:
            result.append(os.path.join(root, f))
        for d in dirs:
            result.append(os.path.join(root, d) + os.sep)
    if not include_root_dir:
        result = [
            os.path.relpath(e, path) + os.sep
            if e.endswith(os.sep)
            else os.path.relpath(e, path)
            for e in result
        ]
    if unixpath:
        result = [e3.os.fs.unixpath(e) for e in result]
    result.sort()
    return result


def echo_to_file(
    filename: str, content: Union[str, List[str]], append: bool = False
) -> None:
    """Output content into a file.

    This function is useful when writing few content to a file for which we
    don't want to keep a file descriptor opened . In other cases, it's more
    efficient to open a file and use the regular python I/O functions.

    :param filename: file to write into
    :param content: string to be written
    :param append: if True append to the file, otherwise overwrite.
    """
    with open(filename, "a+" if append else "w+") as fd:
        if append:
            fd.seek(0, 2)

        if isinstance(content, list):
            for line in content:
                fd.write(line + "\n")
        else:
            fd.write(content)


def find(
    root: str,
    pattern: Optional[str] = None,
    include_dirs: bool = False,
    include_files: bool = True,
    follow_symlinks: bool = False,
) -> List[str]:
    """Find files or directory recursively.

    :param root: directory from which the research start
    :param pattern: glob pattern that files or directories should match in
        order to be included in the final result
    :param include_dirs: if True include directories
    :param include_files: if True include regular files
    :param follow_symlinks: if True include symbolic links

    :return: a list of files
    """
    result = []
    for root, dirs, files in os.walk(root, followlinks=follow_symlinks):
        root = root.replace("\\", "/")
        if include_files:
            for f in files:
                if pattern is None or fnmatch.fnmatch(f, pattern):
                    result.append(root + "/" + f)
        if include_dirs:
            for d in dirs:
                if pattern is None or fnmatch.fnmatch(d, pattern):
                    result.append(root + "/" + d)
    return result


def get_filetree_state(path: str, ignore_hidden: bool = True) -> str:
    """Compute a hash on a filetree to reflect its current state.

    :param path: root path of the file tree to be checked
    :param ignore_hidden: if True (default) then files and directories
        tarting with a dot are ignored.
    :return: a hash as a string

    The function will not report changes in the hash if a file is modified
    and its attributes (size, modification time and mode) are not changed.
    This case is quite uncommon. By ignoring it we can compute efficiently a
    hash representing the state of the file tree without having to read the
    content of all files.
    """

    def compute_state(file_path: str) -> bytes:
        f_stat = os.lstat(file_path)

        state = ":".join(
            [file_path, str(f_stat.st_mode), str(f_stat.st_size), str(f_stat.st_mtime)]
        )
        return state.encode("utf-8")

    path = os.path.abspath(path)
    result = hashlib.sha1()
    if os.path.isdir(path):
        for root, dirs, files in os.walk(path):
            if ignore_hidden:
                ignore_dirs = []
                for index, name in enumerate(dirs):
                    if name.startswith("."):
                        ignore_dirs.append(index)
                ignore_dirs.reverse()
                for index in ignore_dirs:
                    del dirs[index]

            for path in files:
                if ignore_hidden and path.startswith("."):
                    continue

                full_path = os.path.join(root, path)
                result.update(compute_state(full_path))

    else:
        result.update(compute_state(path))
    return result.hexdigest()


def ls(path: Union[str, List[str]], emit_log_record: bool = True) -> List[str]:
    """List files.

    :param path: glob pattern or glob pattern list
    :param emit_log_record: if True, emit a log (debug) record

    :return: a list of filenames

    This function do not raise an error if no file matching the glob pattern
    is encountered. The only consequence is that an empty list is returned.
    """
    if isinstance(path, str):
        path_list = [path]
    else:
        path_list = list(path)

    if emit_log_record:
        logger.debug("ls %s", " ".join(path_list))

    return sorted(itertools.chain.from_iterable(glob.glob(p) for p in path_list))


def mkdir(path: str, mode: int = 0o755, quiet: bool = False) -> None:
    """Create a directory.

    :param path: path to create. If intermediate directories do not exist
        the procedure create them
    :param mode: default is 0755
    :param quiet: whether a log record should be emitted when creating the
        directory
    :raise FSError: if an error occurs

    This function behaves quite like mkdir -p command shell. So if the
    directory already exist no error is raised.
    """
    if os.path.isdir(path):
        return
    else:
        if not quiet:
            logger.debug("mkdir %s (mode=%s)", path, oct(mode))
        try:
            os.makedirs(path, mode)
        except Exception as e:  # defensive code
            if os.path.isdir(path):
                # Take care of cases where in parallel execution environment
                # the directory is created after the initial test on its
                # existence and the call to makedirs
                return
            logger.error(e)
            raise FSError(
                origin="mkdir", message="can't create %s" % path
            ).with_traceback(sys.exc_info()[2])


def mv(source: Union[str, List[str]], target: str) -> None:
    """Move files.

    :param source: a glob pattern
    :param target: target file or directory. If the source resolves as
        several files then target should be a directory

    :raise FSError: if an error occurs
    """

    def move_file(src: str, dst: str) -> None:
        """Reimplementation of shutil.move.

        The implementation follows shutil.move from the standard library.
        The only difference is that we use e3.fs.rm function instead of
        rmtree. This ensure moving a directory with read-only files will
        work.
        """

        def same_file(src: str, dst: str) -> bool:
            if hasattr(os.path, "samefile"):
                try:
                    return os.path.samefile(src, dst)
                except OSError:
                    return False
            return os.path.normcase(os.path.abspath(src)) == os.path.normcase(
                os.path.abspath(dst)
            )

        def basename(path: str) -> str:
            sep = os.path.sep + (os.path.altsep or "")
            return os.path.basename(path.rstrip(sep))

        def destinsrc(src: str, dst: str) -> bool:
            src = os.path.abspath(src)
            dst = os.path.abspath(dst)
            if not src.endswith(os.path.sep):
                src += os.path.sep
            if not dst.endswith(os.path.sep):
                dst += os.path.sep
            return dst.startswith(src)

        real_dst = dst
        if os.path.isdir(dst):
            if same_file(src, dst):
                # We might be on a case insensitive filesystem,
                # perform the rename anyway.
                os.rename(src, dst)
                return

            real_dst = os.path.join(dst, basename(src))
            if os.path.exists(real_dst):
                raise FSError("Destination path '%s' already exists" % real_dst)
        try:
            os.rename(src, real_dst)
        except OSError:
            if os.path.islink(src):
                linkto = os.readlink(src)
                os.symlink(linkto, real_dst)
                os.unlink(src)
            elif os.path.isdir(src):
                if destinsrc(src, dst):
                    raise FSError(
                        "Cannot move a directory '%s' into itself" " '%s'." % (src, dst)
                    )
                shutil.copytree(src, real_dst, symlinks=True)
                rm(src, recursive=True)
            else:
                shutil.copy2(src, real_dst)
                rm(src)
        return

    if isinstance(source, str):
        logger.debug("mv %s %s", source, target)
    else:
        logger.debug("mv %s %s", " ".join(source), target)

    try:
        # Compute file list and number of file to copy
        file_list = ls(source, emit_log_record=False)
        nb_files = len(file_list)

        if nb_files == 0:
            raise FSError(
                origin="mv", message='cannot find files matching "%s"' % source
            )
        elif nb_files == 1:
            source = file_list[0]
            if os.path.isdir(source) and os.path.isdir(target):
                move_file(source, os.path.join(target, os.path.basename(source)))
            else:
                move_file(source, target)
        elif not os.path.isdir(target):
            # More than one file to move but the target is not a directory
            raise FSError("mv", "%s should be a directory" % target)
        else:
            for f in file_list:
                f_dest = os.path.join(target, os.path.basename(f))
                e3.log.debug("mv %s %s", f, f_dest)
                move_file(f, f_dest)
    except Exception as e:
        logger.error(e)
        raise FSError(origin="mv", message=str(e)).with_traceback(sys.exc_info()[2])


def rm(path: Union[str, List[str]], recursive: bool = False, glob: bool = True) -> None:
    """Remove files.

    :param path: a glob pattern, or a list of glob patterns
    :param recursive: if True do a recursive deletion. Default is False
    :param glob: if True globbing pattern expansion is used

    :raise FSError: if an error occurs

    Note that the function will not raise an error is there are no file to
    delete.
    """
    if recursive:
        logger.debug("rm -r %s", str(path))
    else:
        logger.debug("rm %s", str(path))

    # We transform the list into a set in order to remove duplicate files in
    # the list
    if glob:
        file_list = set(ls(path, emit_log_record=False))
    else:
        if isinstance(path, str):
            file_list = {path}
        else:
            file_list = set(path)

    def onerror(func: Callable, error_path: str, exc_info: Tuple) -> None:
        """When shutil.rmtree fail, try again to delete the file.

        :param func: function to call on error
        :param error_path: file or directory to remove
        :param exc_info: exception raised when the first delete attempt was
             made
        """
        del exc_info
        e3.log.debug("error when running %s on %s", func, error_path)

        # First check whether the file we are trying to delete exist. If not
        # the work is already done, no need to continue trying removing it.
        if not os.path.exists(error_path):
            return

        if func in (os.remove, os.unlink):
            # Cannot remove error_path, call chmod and redo an attempt

            # This function is only called when deleting a file inside a
            # directory to remove, it is safe to change the parent directory
            # permission since the parent directory will also be removed.
            os.chmod(os.path.dirname(error_path), 0o700)

            # ??? It seems that this might be needed on windows
            os.chmod(error_path, 0o700)
            e3.os.fs.safe_remove(error_path)

        elif func == os.rmdir:
            # Cannot remove error_path, call chmod and redo an attempt
            os.chmod(error_path, 0o700)

            # Also change the parent directory permission if it will also
            # be removed.
            if recursive and error_path not in file_list:
                # If error_path not in the list of directories to remove it
                # means that we are already in a subdirectory.
                os.chmod(os.path.dirname(error_path), 0o700)
            e3.os.fs.safe_rmdir(error_path)

        elif func in (os.listdir, os.open):
            # Cannot read the directory content, probably a permission issue
            os.chmod(error_path, 0o700)

            # And continue to delete the subdir
            shutil.rmtree(error_path, onerror=onerror)

    for f in file_list:
        try:
            # When calling rmtree or remove, ensure that the string that is
            # passed to this function is unicode on Windows. Otherwise,
            # the non-Unicode API will be used and so we won't be
            # able to remove these files. On Unix don't do that as
            # we got some strange unicode "ascii codec" errors
            # (need some further investigation at some point)
            if sys.platform == "win32":  # unix: no cover
                f = str(f)

            # Note: shutil.rmtree requires its argument to be an actual
            # directory, not a symbolic link to a directory
            if recursive and os.path.isdir(f) and not os.path.islink(f):
                shutil.rmtree(f, onerror=onerror)
            else:
                e3.os.fs.force_remove_file(f)

        except Exception as e:  # defensive code
            logger.error(e)
            raise FSError(
                origin="rm", message="error occurred while removing %s" % f
            ).with_traceback(sys.exc_info()[2])


def splitall(path: str) -> Tuple[str, ...]:
    """Split a path into a list of path components.

    :param path: path to split
    :return: a list of path components
    """
    dirnames = []  # type: List[str]
    while 1:
        head, tail = os.path.split(path)
        if head == path:
            # absolute paths
            # os.path.split('/') -> ('/', '')
            dirnames.append(head)
            break
        elif tail == path:
            # relative paths
            # os.path.split('..') -> ('', '..')
            dirnames.append(tail)
            break
        elif tail == "":
            # ending with a directory separator
            # os.path.split('a/b/c/') -> ('a/b/c', '')
            pass
        else:
            dirnames.append(tail)
        path = head
    return tuple(reversed(dirnames))


VCS_IGNORE_LIST = (
    "RCS",
    "SCCS",
    "CVS",
    "CVS.adm",
    "RCSLOG",
    ".svn",
    ".git",
    ".hg",
    ".bzr",
    ".cvsignore",
    ".gitignore",
    ".gitattributes",
    ".gitmodules",
    ".gitreview",
    ".mailmap",
    ".idea",
)


def sync_tree(
    source: str,
    target: str,
    ignore: Optional[Union[str, Sequence[str]]] = None,
    file_list: Optional[List[str]] = None,
    delete: bool = True,
    preserve_timestamps: bool = True,
    delete_ignore: bool = False,
) -> Tuple[List[str], List[str]]:
    """Synchronize the files and directories between two directories.

    :param source: the directory from where the files and directories
        need to be copied
    :param target: the target directory
    :param ignore: glob pattern or list of files or directories to ignore,
        if the name starts with `/` then only the path is taken into
        account from the root of the source (or target) directory.
        If the ignore value contains a glob pattern, it is taken in account
        only if it doesn't contain a /, since for now the filtering
        is not segmented by '/'.
    :param file_list: list of files to synchronize, if empty synchronize all
        files. Note that if file in the list is a directory then the complete
        content of that directory is included. Note also that ignore list
        takes precedence other file_list.
    :param delete: if True, remove files from target if they do not exist
        in source
    :param preserve_timestamps: if True preserve original timestamps.
        If False updated files get their timestamps set to current time.
    :param delete_ignore: if True files that are explicitely ignored
        are deleted. Note delete should be set to True in that case.
    """
    # Some structure used when walking the trees to be synched
    FilesInfo = namedtuple("FilesInfo", ["rel_path", "source", "target"])

    # The basename in the FileInfo structure is used to compare casing of
    # source and destination.
    FileInfo = namedtuple("FileInfo", ["path", "stat", "basename"])

    # Normalize casing function for path comparison. path_key function
    # return a version of the path that is in lower case for case sensitive
    # and case preserving filesystems. The return value can be used for
    # path comparisons.
    if sys.platform == "win32":

        def path_key(p: str) -> str:
            return p.lower()

    else:

        def path_key(p: str) -> str:
            return p

    # normalize the list of file to synchronize
    norm_file_list = None
    if file_list is not None:
        norm_file_list = [wf.replace("\\", "/").rstrip("/") for wf in file_list]

    # normalize ignore patterns
    if ignore is not None:
        norm_ignore_list = [fn.replace("\\", "/") for fn in ignore]
        abs_ignore_patterns = [fn for fn in norm_ignore_list if fn.startswith("/")]
        rel_ignore_patterns = [fn for fn in norm_ignore_list if not fn.startswith("/")]

    def is_in_ignore_list(p: str) -> bool:
        """Check if a file should be ignored.

        :param p: path relative to source directory (note it starts with a /)

        :return: True if in the list of file to include
        """
        if ignore is None:
            return False

        pk = path_key(p)

        return (
            any(
                f
                for f in abs_ignore_patterns
                if pk == path_key(f) or pk.startswith(path_key(f + "/"))
            )
            or any(
                f
                for f in rel_ignore_patterns
                if pk[1:] == path_key(f) or p.endswith("/" + f)
            )
            or any(
                f
                for f in norm_ignore_list
                if "/" not in f and fnmatch.fnmatch(os.path.basename(p), f)
            )
        )

    def is_in_file_list(p: str) -> bool:
        """Check if a file should be included.

        :param p: path relative to source directory (note it starts with a /)

        :return: True if in the list of file to include
        """
        if file_list is None:
            return True
        if TYPE_CHECKING:
            assert norm_file_list is not None

        pk = path_key(p)

        return any(
            f
            for f in norm_file_list
            if path_key(f) == pk[1:]
            or pk.startswith(path_key("/" + f + "/"))
            or path_key(f).startswith(pk[1:] + "/")
        )

    def isdir(fi: FileInfo) -> bool:
        """Check if a file is a directory.

        :param fi: a FileInfo namedtuple

        :return: True if fi is a directory
        """
        return fi.stat is not None and stat.S_ISDIR(fi.stat.st_mode)

    def islink(fi: FileInfo) -> bool:
        """Check if a file is a link.

        :param fi: a FileInfo namedtuple

        :return: True if fi is a symbolic link
        """
        return fi.stat is not None and stat.S_ISLNK(fi.stat.st_mode)

    def isfile(fi: FileInfo) -> bool:
        """Check if a file is a regular file.

        :param fi: a FileInfo namedtuple
        :return: True if fi is a regular file
        """
        return fi.stat is not None and stat.S_ISREG(fi.stat.st_mode)

    def cmp_files(src: FileInfo, dst: FileInfo) -> bool:
        """Fast compare two files."""
        bufsize = 8 * 1024
        with open(src.path, "rb") as fp1, open(dst.path, "rb") as fp2:
            while True:
                b1 = fp1.read(bufsize)
                b2 = fp2.read(bufsize)
                if b1 != b2:
                    return False

                if len(b1) < bufsize:
                    return True

    def need_update(src: FileInfo, dst: FileInfo) -> bool:
        """Check if dst file should updated.

        :param src: the source FileInfo object
        :param dst: the target FileInfo object

        :return: True if we should update dst
        """
        # when not preserving timestamps we cannot rely on the timestamps to
        # check if a file is up-to-date. In that case do a full content
        # comparison as last check.
        return (
            dst.stat is None
            or stat.S_IFMT(src.stat.st_mode) != stat.S_IFMT(dst.stat.st_mode)
            or (
                preserve_timestamps
                and abs(src.stat.st_mtime - dst.stat.st_mtime) > 0.001
            )
            or src.stat.st_size != dst.stat.st_size
            or (not preserve_timestamps and isfile(src) and not cmp_files(src, dst))
            or src.basename != dst.basename
        )

    def copystat(src: FileInfo, dst: FileInfo) -> None:
        """Update attribute of dst file with src attributes.

        :param src: the source FileInfo object
        :param dst: the target FileInfo object
        """
        if islink(src):  # windows: no cover
            mode = stat.S_IMODE(src.stat.st_mode)
            if hasattr(os, "lchmod"):
                os.lchmod(dst.path, mode)

            if hasattr(os, "lchflags") and hasattr(src.stat, "st_flags"):
                try:
                    os.lchflags(dst.path, src.stat.st_flags)
                except OSError as why:  # defensive code
                    import errno

                    if (
                        not hasattr(errno, "EOPNOTSUPP")
                        or why.errno != errno.EOPNOTSUPP
                    ):
                        raise
        else:
            mode = stat.S_IMODE(src.stat.st_mode)
            if hasattr(os, "utime"):
                if preserve_timestamps:
                    os.utime(dst.path, (src.stat.st_atime, src.stat.st_mtime))
                else:
                    os.utime(dst.path, None)
            if hasattr(os, "chmod"):
                os.chmod(dst.path, mode)
            if hasattr(os, "chflags") and hasattr(src.stat, "st_flags"):
                try:
                    os.chflags(dst.path, src.stat.st_flags)
                except OSError as why:  # defensive code
                    import errno

                    if (
                        not hasattr(errno, "EOPNOTSUPP")
                        or why.errno != errno.EOPNOTSUPP
                    ):
                        raise

    def safe_copy(src: FileInfo, dst: FileInfo) -> None:
        """Copy src file into dst preserving all attributes.

        :param src: the source FileInfo object
        :param dst: the target FileInfo object
        """
        if islink(src):  # windows: no cover
            linkto = os.readlink(src.path)
            if not islink(dst) or os.readlink(dst.path) != linkto:
                if dst.stat is not None:
                    rm(dst.path, recursive=True, glob=False)
                os.symlink(linkto, dst.path)
            copystat(src, dst)
        else:
            if isdir(dst):
                # dst directory will be replaced by a file having the same
                # content as 'src'
                rm(dst.path, recursive=True, glob=False)
            elif islink(dst):
                # dst symlink will be replaced by a file having the same
                #  content as 'src'
                rm(dst.path, recursive=False, glob=False)

            try:
                if dst.basename != src.basename:
                    rm(dst.path, glob=False)
                    dst = FileInfo(
                        os.path.join(os.path.dirname(dst.path), src.basename),
                        None,
                        src.basename,
                    )

                with open(src.path, "rb") as fsrc:
                    with open(dst.path, "wb") as fdst:
                        shutil.copyfileobj(fsrc, fdst)  # type: ignore
            except OSError:
                rm(dst.path, glob=False)
                with open(src.path, "rb") as fsrc:
                    with open(dst.path, "wb") as fdst:
                        shutil.copyfileobj(fsrc, fdst)  # type: ignore
            copystat(src, dst)

    def safe_mkdir(src: FileInfo, dst: FileInfo) -> None:
        """Create a directory modifying parent directory permissions if needed.

        :param dst: directory to create
        """
        if isfile(dst) or islink(dst):
            rm(dst.path, glob=False)

        try:
            if isdir(dst):
                if dst.basename != src.basename:
                    os.rename(
                        dst.path, os.path.join(os.path.dirname(dst.path), src.basename)
                    )
            else:
                os.makedirs(dst.path)
        except OSError:
            # in case of error to change parent directory
            # permissions. The permissions will be then
            # set correctly at the end of rsync.
            e3.os.fs.chmod("a+wx", os.path.dirname(dst.path))
            os.makedirs(dst.path)

    def walk(
        root_dir: str, target_root_dir: str, entry: Optional[FilesInfo] = None
    ) -> Iterable[FilesInfo]:
        """Walk through source and target file trees.

        :param root_dir: path to source tree
        :param target_root_dir: path to target tree
        :param entry: a FilesInfo object (used internally for the recursion)

        :return: an iterator that iterate other the relevant FilesInfo object
        """
        if entry is None:
            target_stat = None
            if os.path.exists(target_root_dir):
                target_stat = os.lstat(target_root_dir)

            entry = FilesInfo(
                "",
                FileInfo(root_dir, os.lstat(root_dir), ""),
                FileInfo(target_root_dir, target_stat, ""),
            )
            yield entry

        try:
            source_names = {path_key(k): k for k in os.listdir(entry.source.path)}
        except Exception:  # defensive code
            e3.log.debug("cannot get sources list", exc_info=True)
            # Don't crash in case a source directory cannot be read
            return

        target_names = {}
        if isdir(entry.target):
            try:
                target_names = {path_key(k): k for k in os.listdir(entry.target.path)}
            except Exception:
                e3.log.debug("cannot get targets list", exc_info=True)
                target_names = {}

        all_names = set(source_names.keys()) | set(target_names.keys())

        result = []
        for name in all_names:
            rel_path = f"{entry.rel_path}/{name}"

            source_full_path = os.path.join(
                entry.source.path, source_names.get(name, name)
            )
            target_full_path = os.path.join(
                entry.target.path, target_names.get(name, name)
            )
            source_stat = None
            target_stat = None

            if name in source_names:
                source_stat = os.lstat(source_full_path)

            source_file = FileInfo(
                source_full_path, source_stat, os.path.basename(source_full_path)
            )

            if name in target_names:
                target_stat = os.lstat(target_full_path)

            target_file = FileInfo(
                target_full_path, target_stat, os.path.basename(target_full_path)
            )

            result.append(FilesInfo(rel_path, source_file, target_file))

        for el in result:
            if is_in_ignore_list(el.rel_path):
                logger.debug("ignore %s", el.rel_path)
                if delete_ignore:
                    yield FilesInfo(
                        el.rel_path,
                        FileInfo(el.source.path, None, el.source.basename),
                        el.target,
                    )
            elif is_in_file_list(el.rel_path):
                yield el
                if isdir(el.source):
                    yield from walk(root_dir, target_root_dir, el)
            else:
                yield FilesInfo(
                    el.rel_path,
                    FileInfo(el.source.path, None, el.source.basename),
                    el.target,
                )

    source_top = os.path.normpath(source).rstrip(os.path.sep)
    target_top = os.path.normpath(target).rstrip(os.path.sep)
    copystat_dir_list = []

    logger.debug(
        "sync_tree %s -> %s [delete=%s, preserve_stmp=%s]",
        source,
        target,
        delete,
        preserve_timestamps,
    )

    if not os.path.exists(source):
        raise FSError(origin="sync_tree", message="%s does not exist" % source)

    # Keep track of deleted and updated files
    deleted_list: List[str] = []
    updated_list: List[str] = []

    for wf in walk(source_top, target_top):
        if wf.source.stat is None and wf.target.stat is not None:
            # Entry that exist only in the target file tree. Check if we
            # should delete it
            if delete:
                rm(wf.target.path, recursive=True, glob=False)
                deleted_list.append(wf.target.path)
        else:
            # At this stage we have an element to synchronize in
            # the source tree.
            if need_update(wf.source, wf.target):
                if isfile(wf.source) or islink(wf.source):
                    safe_copy(wf.source, wf.target)
                    updated_list.append(wf.target.path)
                elif isdir(wf.source):
                    safe_mkdir(wf.source, wf.target)
                    updated_list.append(wf.target.path)
                    copystat_dir_list.append((wf.source, wf.target))

    # Adjust directory permissions once all files have been copied
    for d in copystat_dir_list:
        copystat(d[0], d[1])

    return updated_list, deleted_list


def extension(path: str) -> str:
    """Return the extension of a given filename.

    Contrary to os.path.splitext which returns .gz, the function will return
    .tar.gz if the file is FILENAME.tar.gz.

    :param path: a path
    :return: an extension
    """
    root, ext = os.path.splitext(path)
    _, ext2 = os.path.splitext(root)
    if ext2 == ".tar":
        return ext2 + ext
    else:
        return ext



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import List, Optional, Union


class E3Error(Exception):
    """Exception raised by functions defined in E3."""

    def __init__(self, message: Union[str, List[str]], origin: Optional[str] = None):
        """Initialize an E3Error.

        E3Error can store several messages and thus be used to propagate them.

        :param message: the exception message
        :param origin: the name of the function, class, or module having raised
            the exception
        """
        super().__init__(message, origin)
        self.origin = origin
        self.messages = []
        if message is not None:
            if isinstance(message, str):
                self.messages.append(message)
            else:
                self.messages.extend(message)

    def __iadd__(self, other: Union[str, List[str], E3Error]) -> E3Error:
        """Add messages to the current instance.

        :param other: a message or an E3Error instance
        """
        if isinstance(other, E3Error):
            self.messages.extend(other.messages)
        elif isinstance(other, str):
            self.messages.append(other)
        else:
            self.messages.extend(other)
        return self

    def __str__(self) -> str:
        if self.messages:
            error_msg = self.messages[-1]
        else:
            error_msg = self.__class__.__name__
        if self.origin:
            return f"{self.origin}: {error_msg}\n"
        else:
            return error_msg



from __future__ import annotations

from typing import TYPE_CHECKING
import yaml

from e3.env import BaseEnv

if TYPE_CHECKING:
    from typing import Any, Dict, List, Optional


class Host(BaseEnv):
    """Represent an host from the host database.

    See e3.env.BaseEnv
    """

    def __init__(
        self, hostname: str, platform: str, version: str, **kwargs: Any
    ) -> None:
        """Initialize an host entry.

        :param hostname: host name
        :param platform: platform name (see e3.platform)
        :param version: platform version (usually OS version)
        :param kwargs: additional user defined data. each key from the data
            dict is accessible like a regular attribute.
        """
        BaseEnv.__init__(self)
        self.set_build(name=str(platform), version=str(version), machine=str(hostname))
        self._instance.update(kwargs)


class HostDB:
    """Host database.

    :ivar hosts: dict indexed by host name
    """

    def __init__(self, filename: Optional[str] = None):
        """Initialize a host database.

        :param filename: if not None, initialize the database from a yaml
            file. See HostDB.load_yaml_db method for details about the expected
            format
        """
        self.hosts: Dict[str, Host] = {}

        if filename is not None:
            self.load_yaml_db(filename)

    @property
    def hostnames(self) -> List[str]:
        """Return the current list of host names.

        :return: a list of hostnames
        """
        return list(self.hosts.keys())

    def add_host(self, hostname: str, platform: str, version: str, **data: Any) -> None:
        """Add/Update a host entry.

        :param hostname: host name
        :param platform: platform name (see e3.platform)
        :param version: platform/OS version
        :param data: additional host information
        """
        self.hosts[hostname] = Host(hostname, platform, version, **data)

    def load_yaml_db(self, filename: str) -> None:
        """Load a yaml configuration file.

        The yaml file should be a dictionaty indexed by host name. Each entry
        should be a dictionary containing at least the following two keys:

        * build_platform: the platform name
        * build_os_version: the platform/os version

        Additional keys for a host entry will be considered as additional data

        :param filename: path the yaml file
        """
        with open(filename) as fd:
            content = yaml.safe_load(fd)

        for hostname, hostinfo in content.items():
            result = {}
            for key in hostinfo:
                if key not in ("build_platform", "build_os_version"):
                    result[key] = hostinfo[key]
            platform = hostinfo["build_platform"]
            version = hostinfo["build_os_version"]
            self.add_host(hostname, platform, version, **result)

    def __getitem__(self, key: str) -> Host:
        return self.hosts[key]

    def get(self, key: str, default: Optional[Host] = None) -> Optional[Host]:
        """Return the Host named ``key`` of ``default``.

        :param key: host name
        :param default: default value to return if not found
        """
        return self.hosts.get(key, default)



import os

from e3.anod.driver import AnodDriver
from e3.anod.helper import Configure, Make, text_replace
from e3.anod.sandbox import SandBox
from e3.anod.spec import Anod
from e3.env import BaseEnv


def test_make():
    class AnodMake(Anod):
        def shell(self, *cmd, **kwargs):
            """Mock for Anod.shell that does not spawn processes."""
            return (cmd, kwargs)

        @Anod.primitive()
        def build(self):
            m1 = Make(self, makefile="/tmp/makefile")
            m1.set_var("prefix", "/foo")
            m2 = Make(self, jobs=2)
            m2.set_default_target("install")
            m2.set_var("profiles", ["dev", "prod"])
            return (
                m1.cmdline()["cmd"],
                m1()[0],
                m1(exec_dir="/foo", timeout=2)[1],
                m2.cmdline()["cmd"],
                m2.cmdline(["clean", "install"])["cmd"],
                m2.cmdline("all")["cmd"],
            )

    Anod.sandbox = SandBox()
    Anod.sandbox.root_dir = os.getcwd()
    Anod.sandbox.create_dirs()

    am = AnodMake(qualifier="", kind="build", jobs=10)
    AnodDriver(anod_instance=am, store=None).activate(Anod.sandbox, None)
    am.build_space.create()
    assert am.build() == (
        ["make", "-f", "/tmp/makefile", "-j", "10", "prefix=/foo"],
        ("make", "-f", "/tmp/makefile", "-j", "10", "prefix=/foo"),
        {"cwd": "/foo", "timeout": 2},
        ["make", "-j", "2", "profiles=dev prod", "install"],
        ["make", "-j", "2", "profiles=dev prod", "clean", "install"],
        ["make", "-j", "2", "profiles=dev prod", "all"],
    )


def test_configure():
    class AnodConf(Anod):
        @Anod.primitive()
        def build(self):
            c = Configure(self)
            return c.cmdline()

    Anod.sandbox = SandBox()
    Anod.sandbox.root_dir = os.getcwd()
    Anod.sandbox.create_dirs()

    ac = AnodConf(qualifier="", kind="build", jobs=10)
    AnodDriver(anod_instance=ac, store=None).activate(Anod.sandbox, None)
    ac.build_space.create()

    # Configure() can add $CONFIG_SHELL in the command line
    # Check that the two other arguments are as expected
    assert ac.build()["cmd"][-2:] == [
        "../src/configure",
        "--build=%s" % ac.env.build.triplet,
    ]

    # Check with canadian env

    canadian_env = BaseEnv()
    canadian_env.set_build("x86-windows")
    canadian_env.set_host("x86-linux")
    canadian_env.set_target("arm-elf")
    assert canadian_env.is_canadian

    ac2 = AnodConf(qualifier="", kind="build", jobs=10, env=canadian_env)
    AnodDriver(anod_instance=ac2, store=None).activate(Anod.sandbox, None)
    ac2.build_space.create()

    ac2_cmd = ac2.build()["cmd"]
    assert "--build=i686-pc-mingw32" in ac2_cmd
    assert "--host=i686-pc-linux-gnu" in ac2_cmd
    assert "--target=arm-eabi" in ac2_cmd

    # Check with cross env

    cross_env = BaseEnv()
    cross_env.set_target("arm-elf")

    ac3 = AnodConf(qualifier="", kind="build", jobs=10, env=cross_env)
    AnodDriver(anod_instance=ac3, store=None).activate(Anod.sandbox, None)
    ac3.build_space.create()

    assert "--target=arm-eabi" in ac3.build()["cmd"]


def test_configure_opts():
    """Check configure options."""

    class AnodConf(Anod):
        def shell(self, *cmd, **kwargs):
            """Mock for Anod.shell that does not spawn processes."""
            return (cmd, kwargs)

        @Anod.primitive()
        def build(self):
            c = Configure(self)
            c.add("--with-opt")
            c.add_env("OPT", "VAL")
            return [c.cmdline(), c()]

    os.environ["CONFIG_SHELL"] = "ksh"

    Anod.sandbox = SandBox()
    Anod.sandbox.root_dir = os.getcwd()
    Anod.sandbox.create_dirs()

    ac = AnodConf(qualifier="", kind="build", jobs=10)
    AnodDriver(anod_instance=ac, store=None).activate(Anod.sandbox, None)
    ac.build_space.create()

    result = ac.build()

    assert result[0]["cmd"][:-1] == ["ksh", "../src/configure", "--with-opt"]
    assert result[0]["options"]["env"] == {"OPT": "VAL"}

    assert result[1][0][:-1] == tuple(result[0]["cmd"][:-1])
    assert result[1][1]["env"] == result[0]["options"]["env"]


def test_text_replace():
    with open("myfile", "w") as f:
        f.write("what who when")
    text_replace("myfile", [(b"who", b"replaced")])
    with open("myfile") as f:
        assert f.read() == "what replaced when"



import os
import subprocess
import sys

from e3.anod.driver import AnodDriver
from e3.anod.error import AnodError, SpecError
from e3.anod.sandbox import SandBox
from e3.anod.spec import Anod, __version__, check_api_version, has_primitive

import pytest


def test_simple_spec():
    class Simple(Anod):

        test_qualifier_format = (("with_bar", False),)

        build_source_list = [
            Anod.Source("foo-src", publish=False),
            Anod.Source("bar-src", publish=True),
        ]

        @property
        def test_source_list(self):
            result = [Anod.Source("foo-test-src", publish=False)]
            if self.parsed_qualifier.get("with_bar"):
                result.append(Anod.Source("bar-test-src", publish=False))
            return result

    simple_build = Simple("", kind="build")
    assert len(simple_build.build_source_list) == 2

    simple_test = Simple("", kind="test")
    assert len(simple_test.test_source_list) == 1

    simple_test_with_bar = Simple("with_bar=true", kind="test")
    assert len(simple_test_with_bar.test_source_list) == 2


def test_spec_buildvars():
    """Build vars are used by the driver and not visible in deps."""

    class MySpec(Anod):

        build_deps = [Anod.BuildVar("key", "value")]

        @Anod.primitive()
        def build(self):
            pass

    ms = MySpec("", kind="build")
    assert len(ms.deps) == 0


def test_spec_wrong_dep():
    """Check exception message when wrong dependency is set."""
    with pytest.raises(SpecError) as err:
        Anod.Dependency("foo", require="invalid")

    assert (
        "require should be build_tree, installation or source_pkg not "
        "invalid" in str(err)
    )


def test_primitive():
    class NoPrimitive(Anod):
        def build(self):
            return 2

    no_primitive = NoPrimitive("", "build")
    assert has_primitive(no_primitive, "build") is False

    class WithPrimitive(Anod):

        build_qualifier_format = (("error", False),)

        package = Anod.Package(prefix="mypackage", version=lambda: "42")

        @Anod.primitive()
        def build(self):
            if "error" in self.parsed_qualifier:
                raise ValueError(self.parsed_qualifier["error"])
            elif "error2" in self.parsed_qualifier:
                self.shell(sys.executable, "-c", "import sys; sys.exit(2)")
            else:
                hello = self.shell(
                    sys.executable, "-c", 'print("world")', output=subprocess.PIPE
                )
                return hello.out.strip()

    with_primitive = WithPrimitive("", "build")
    with_primitive2 = WithPrimitive("error=foobar", "build")
    with_primitive3 = WithPrimitive("error2", "build")
    with_primitive4 = WithPrimitive("error3", "build")

    Anod.sandbox = SandBox()
    Anod.sandbox.root_dir = os.getcwd()
    Anod.sandbox.spec_dir = os.path.join(os.path.dirname(__file__), "data")
    Anod.sandbox.create_dirs()
    # Activate the logging
    AnodDriver(anod_instance=with_primitive, store=None).activate(Anod.sandbox, None)
    AnodDriver(anod_instance=with_primitive2, store=None).activate(Anod.sandbox, None)
    AnodDriver(anod_instance=with_primitive3, store=None).activate(Anod.sandbox, None)
    AnodDriver(anod_instance=with_primitive4, store=None)  # don't activate

    with_primitive.build_space.create()

    assert has_primitive(with_primitive, "build") is True
    assert with_primitive.build() == "world"

    with_primitive2.build_space.create()

    with pytest.raises(AnodError) as err:
        with_primitive2.build()
    assert "foobar" in str(err.value)

    assert with_primitive2.package.name.startswith("mypackage")

    # Check __getitem__
    # PKG_DIR returns the path to the pkg directory
    assert with_primitive2["PKG_DIR"].endswith("pkg")

    with_primitive3.build_space.create()
    with pytest.raises(AnodError) as err:
        with_primitive3.build()
    assert "build fails" in str(err.value)


def test_api_version():
    # __version__ is supported
    check_api_version(__version__)

    with pytest.raises(AnodError):
        check_api_version("0.0")



import os
import sys
import textwrap
import time

import e3.fs
import e3.os.fs
import e3.os.process

import pytest

try:
    import psutil
except ImportError:
    psutil = None


def test_run_shebang(caplog):
    """Verify that the parse shebang option works."""
    prog_filename = os.path.join(os.getcwd(), "prog")
    with open(prog_filename, "wb") as f:
        f.write(b"#!/usr/bin/env python\n")
        f.write(b"import sys\n")
        f.write(b'print("running %s" % sys.argv[1])\n')
    e3.os.fs.chmod("a+x", prog_filename)
    p = e3.os.process.Run([prog_filename, "atest"], parse_shebang=True)
    assert p.out.replace("\r", "") == "running atest\n"

    # Create a shebang spawning a file that does not exist
    with open(prog_filename, "wb") as f:
        f.write(b"#!doesnot exist\n")
        f.write(b'print("running python prog")\n')

    e3.os.fs.chmod("a+x", prog_filename)
    with pytest.raises(OSError) as err:
        e3.os.process.Run([prog_filename], parse_shebang=True)
    assert "doesnot" in str(err)
    assert "doesnot exist" in caplog.text


def test_split_err_out():
    """Split err and out to distinct pipes."""
    p = e3.os.process.Run(
        [
            sys.executable,
            "-c",
            "import sys; sys.stdout.write('stdout'); sys.stderr.write('stderr')",
        ],
        output=e3.os.process.PIPE,
        error=e3.os.process.PIPE,
    )
    assert p.out == "stdout"
    assert p.err == "stderr"


def test_non_utf8_out():
    """Test that we can get an output for a process not emitting utf-8."""
    p = e3.os.process.Run(
        [sys.executable, "-c", "import sys; sys.stdout.buffer.write(b'\\xff\\xff')"]
    )
    assert p.out == "\\xff\\xff"


def test_rlimit():
    """rlimit kill the child process after a timeout."""

    def run_test():
        p = e3.os.process.Run(
            [
                sys.executable,
                "-c",
                "print('hello'); import sys; sys.stdout.flush(); "
                "import time; time.sleep(10); print('world')",
            ],
            timeout=1,
        )
        assert "hello" in p.out
        assert "world" not in p.out

    run_test()
    if sys.platform == "win32":
        # On Windows make sure that rlimit works when
        # setting the build environment to 64bit windows
        e = e3.env.Env()
        e.store()
        e.set_build("x86_64-windows")
        run_test()
        e.set_build("x86_64-windows64")
        run_test()
        e.restore()


def test_not_found():
    with pytest.raises(OSError) as err:
        e3.os.process.Run(["e3-bin-not-found"])
    assert "e3-bin-not-found not found" in str(err.value)

    with pytest.raises(OSError) as err:
        e3.os.process.Run(["e3-bin-not-found"], parse_shebang=True)
    assert "e3-bin-not-found not found" in str(err.value)

    with pytest.raises(OSError) as err:
        e3.os.process.Run(
            [
                [
                    sys.executable,
                    "-c",
                    'print("a "); import time; time.sleep(10); print("test")',
                ],
                ["e3-bin-not-found2"],
            ]
        )
    assert "e3-bin-not-found2 not found" in str(err.value)


def test_enable_commands_handler():
    log_file = "cmds.log"
    h = e3.os.process.enable_commands_handler(log_file)
    try:
        e3.os.process.Run([sys.executable, "-c", 'print("dummy")'])
        e3.os.process.Run([sys.executable, "-c", 'print("dummy2")'])
    finally:
        e3.os.process.disable_commands_handler(h)

    with open(log_file, "rb") as fd:
        lines = fd.readlines()
    assert len(lines) == 2


@pytest.mark.xfail(sys.platform != "win32", reason="unix implem not complete")
def test_wait_for_processes():
    for v in (1, 2):
        with open("p%d.py" % v, "w") as f:
            f.write(
                "import os\n"
                "while True:\n"
                '    if os.path.exists("end%d"): break\n'
                'print("process%d")\n' % (v, v)
            )

    p1 = e3.os.process.Run([sys.executable, "p1.py"], bg=True)
    p2 = e3.os.process.Run([sys.executable, "p2.py"], bg=True)

    process_list = [p1, p2]
    p3 = e3.os.process.Run(
        [
            sys.executable,
            "-c",
            "from e3.os.fs import touch;"
            "from time import sleep;"
            "sleep(0.2);"
            'touch("end1");'
            "sleep(0.2);"
            'touch("end2")',
        ],
        bg=True,
    )
    for _ in range(2):
        result = e3.os.process.wait_for_processes(process_list, 2)
        if result is not None:
            del process_list[result]
    assert len(process_list) == 0

    assert p1.status == 0
    assert p1.out.strip() == "process1"
    assert p2.status == 0
    assert p2.out.strip() == "process2"

    p3.wait()

    assert e3.os.process.wait_for_processes([], 10) is None


def test_run_pipe():
    cmd_left = [sys.executable, "-c", 'print("dummy")']
    cmd_right = [
        sys.executable,
        "-c",
        'import sys; print(sys.stdin.read().replace("y", "ies"))',
    ]
    p = e3.os.process.Run([cmd_left, cmd_right])
    assert p.status == 0
    assert p.out.strip() == "dummies"

    with open("dummy", "w") as f:
        f.write("dummy")
    p = e3.os.process.Run(cmd_right, input="dummy")
    assert p.status == 0
    assert p.out.strip() == "dummies"


def test_command_line_image():
    result = e3.os.process.command_line_image(["echo", ""])
    assert result == "echo ''"
    result = e3.os.process.command_line_image([["echo", "dummy"], ["grep", "m"]])
    assert result == "echo dummy | grep m"


def test_poll():
    result = e3.os.process.Run(
        [sys.executable, "-c", 'import time; time.sleep(1); print("process")'], bg=True
    )

    assert result.poll() is None
    time.sleep(2)
    assert result.poll() == 0
    assert result.out.strip() == "process"

    # check that subsequent calls to poll or wait do not crash or alter the
    # result
    assert result.poll() == 0
    assert result.wait() == 0
    assert result.out.strip() == "process"


def test_file_redirection():
    p_out = "p.out"
    result = e3.os.process.Run(
        [sys.executable, "-c", 'print("dummy")'],
        input=None,
        output=p_out,
        error=e3.os.process.STDOUT,
    )
    with open(p_out) as fd:
        content = fd.read().strip()
    assert result.status == 0
    assert content == "dummy"


def test_output_append():
    p_out = "p.out"
    e3.os.process.Run([sys.executable, "-c", 'print("line1")'], output=p_out)
    e3.os.process.Run([sys.executable, "-c", 'print("line2")'], output="+" + p_out)
    with open(p_out) as fd:
        content = fd.read().strip()
    assert content == "line1\nline2"


def test_pipe_input():
    p = e3.os.process.Run(
        [sys.executable, "-c", "import sys; print(sys.stdin.read())"], input="|dummy"
    )
    assert p.out.strip() == "dummy"


def test_is_running():
    p = e3.os.process.Run([sys.executable, "-c", "import time; time.sleep(1)"], bg=True)
    assert e3.os.process.is_running(p.pid)
    p.kill(recursive=False)

    # On windows we don't want to wait as otherwise pid will be reused
    # Note also that the semantic is slightly different between Unix
    # and Windows. is_running will report false on Windows once the
    # process is in a waitable state.
    if sys.platform != "win32":
        p.wait()
    assert not e3.os.process.is_running(p.pid)

    p.wait()


@pytest.mark.skipif(psutil is None, reason="require psutil")
def test_is_running_non_existant():
    """Call is_running on non-existing process."""
    pid_list = psutil.pids()
    pid_list.sort()

    # Try to found a non existing process
    for a in range(1, 1000):
        running = e3.os.process.is_running(pid_list[-1] + a)
        if not running:
            break
    assert not running, "could not find non existing process"


@pytest.mark.xfail(reason="unstable test, p.status can be 0")
def test_interrupt():
    t0 = time.time()
    p = e3.os.process.Run(
        [sys.executable, "-c", "import time; time.sleep(30)"], bg=True
    )
    time.sleep(0.5)  # Make sure the process had the time to start
    p.interrupt()
    t1 = time.time()
    assert t1 - t0 < 2, "process not interrupted after 2s?"

    p.wait()
    assert p.status != 0


@pytest.mark.skipif(psutil is None, reason="require psutil")
def test_kill_process_tree():

    is_appveyor_test = bool(os.environ.get("APPVEYOR"))
    wait_timeout = 3
    if is_appveyor_test:
        wait_timeout *= 3
    p1 = e3.os.process.Run(
        [sys.executable, "-c", "import time; time.sleep(10); import sys; sys.exit(2)"],
        bg=True,
    )
    e3.os.process.kill_process_tree(p1.pid, timeout=wait_timeout)
    assert p1.status != 2
    assert not p1.is_running()

    def get_one_child(idx):
        pid_file = "child_pid_%d" % idx
        gen_prog_name = "child_prog_%d" % idx
        prog = textwrap.dedent(
            """\
            import e3.os.process, os, sys, time
            child_cmd = "import os, time;"
            child_cmd += "f = open('%s', 'w');"
            child_cmd += "f.write(str(os.getpid()));"
            child_cmd += "f.close();"
            child_cmd += "time.sleep(60);"
            e3.os.process.Run([sys.executable, '-c', child_cmd])
            time.sleep(60)
            """
            % pid_file
        )

        with open(gen_prog_name, "w") as f:
            f.write(prog)

        parent_process = e3.os.process.Run([sys.executable, gen_prog_name], bg=True)
        for _ in range(0, 100):
            try:
                with open(pid_file) as f:
                    child_pid = f.read()
                    if child_pid:
                        break
            except OSError:
                pass
            time.sleep(0.1)
        e3.fs.rm(gen_prog_name)
        e3.fs.rm(pid_file)
        child_process = psutil.Process(int(child_pid))
        assert parent_process.children()[0].pid == int(child_pid)
        return parent_process, child_process

    p2, p2_child = get_one_child(1)
    e3.os.process.kill_process_tree(p2.pid, timeout=wait_timeout)

    assert not p2.is_running()
    assert not p2_child.is_running()

    p3, p3_child = get_one_child(2)
    p3.kill(timeout=wait_timeout)

    assert not p3.is_running()
    assert not p3_child.is_running()

    # killing it more than once should also work
    assert e3.os.process.kill_process_tree(p3.pid) is True
    assert e3.os.process.kill_process_tree(p3.internal) is True


def test_run_with_env():
    os.environ["EXT_VAR"] = "bar"
    cmd = (
        [
            sys.executable,
            "-c",
            'import os; print(os.environ.get("TEST_RUN_VAR")'
            ' + os.environ.get("EXT_VAR", "world"))',
        ],
    )
    p1 = e3.os.process.Run(cmd, env={"TEST_RUN_VAR": "foo"}, ignore_environ=False)
    assert p1.out.strip() == "foobar"

    p1 = e3.os.process.Run(cmd, env={"TEST_RUN_VAR": "hello"}, ignore_environ=True)
    assert p1.out.strip() == "helloworld"


def test_no_rlimit(caplog):
    fake_rlimit = e3.os.process.get_rlimit(platform="null")
    old_get_rlimit = e3.os.process.get_rlimit
    e3.os.process.get_rlimit = lambda: fake_rlimit  # type: ignore

    try:
        p1 = e3.os.process.Run([sys.executable, "-c", "print(1)"], timeout=2)
        assert p1.out.strip() == "1"
        assert "cannot find rlimit" in caplog.text
    finally:
        e3.os.process.get_rlimit = old_get_rlimit


@pytest.mark.skipif(sys.platform != "win32", reason="windows specific test")
def test_shell_override():
    """Unix shell shebang handling.

    On windows we ensure that /bin/bash /bin/sh shebangs are replaced by
    SHELL env var.
    """
    work_dir = os.getcwd()
    os.environ["SHELL"] = sys.executable
    test_file_path = os.path.join(work_dir, "shebang_test.sh")
    with open(test_file_path, "w") as fd:
        fd.write("#!/bin/bash\nimport sys; print(sys.executable)\n")
    p = e3.os.process.Run([test_file_path], parse_shebang=True)
    assert p.out.strip() == sys.executable



"""Read e3 config file."""
from __future__ import annotations
from dataclasses import fields, dataclass

from typing import TYPE_CHECKING, get_type_hints, ClassVar

try:
    from typeguard import check_type

    CONFIG_CHECK_TYPE = True
except ImportError:  # defensive code
    CONFIG_CHECK_TYPE = False


import logging
import os

if TYPE_CHECKING:
    from typing import Type, TypeVar

    T = TypeVar("T", bound="ConfigSection")


if "E3_CONFIG" in os.environ:
    KNOWN_CONFIG_FILES = [os.environ["E3_CONFIG"]]
else:
    KNOWN_CONFIG_FILES = [
        os.path.join(
            os.environ.get("XDG_CONFIG_HOME", os.path.expanduser("~/.config")),
            "e3.toml",
        ),
        os.path.expanduser("~/e3.toml"),
    ]


@dataclass
class ConfigSection:
    title: ClassVar[str]

    @classmethod
    def load(cls: Type[T]) -> T:
        """Load a section of the configuration file.

        To load a new section, subclass ConfigSection and document the
        fields that you expect to parse, e.g.

        @dataclass
        class MyConfig(ConfigSection):
            title = "my_config_subsection"
            option : str = "default value"

        my_config = MyConfig.load()

        If the package typeguard is installed type defined in the metaclass
        will be verified.
        """
        schema = get_type_hints(cls)
        cls_fields = {f.name: schema[f.name] for f in fields(cls) if f.name != "title"}
        kwargs = {}

        for k, v in Config.load_section(cls.title).items():
            if k in cls_fields:
                ftype = cls_fields[k]
                try:
                    if CONFIG_CHECK_TYPE:
                        check_type(f"{cls.title}.{k}", v, ftype)
                except TypeError as err:
                    logging.error(str(err))
                else:
                    kwargs[k] = v

        return cls(**kwargs)  # type: ignore


class Config:
    """Load e3 configuration file and validate each section.

    This class expose the .load_section(<section>, <dataclass>) method
    that can be used by ConfigSection instance corresponding to the loaded
    configuration section after validation.

    Note that without the tomlkit package the configuration is not read.
    """

    data: ClassVar[dict] = {}

    @classmethod
    def load_section(cls, section: str) -> dict:
        """Load a configuration section content.

        :param section: if contains "." nested subsection will be found. For
            instance "log.fmt" will return the section:

            [log]
              [log.fmt]
        :return: the configuration dict
        """
        if not cls.data:
            cls.load()

        subsections = section.split(".")
        result = cls.data
        for subsection in subsections:
            result = result.get(subsection, {})

        return result

    @classmethod
    def load_file(cls, filename: str) -> None:
        """Load the configuration file(s).

        Note that this method is automatically loaded the first time .get()
        is called.

        :param filename: configuration file to load
        """
        try:
            from tomlkit import parse
            from tomlkit.exceptions import TOMLKitError
        except ImportError:  # defensive code
            logging.error(f"cannot load {filename} (cannot import tomlkit)")
        else:
            with open(filename) as f:
                try:
                    cls.data.update(parse(f.read()))
                except TOMLKitError as e:
                    logging.error(str(e))

    @classmethod
    def load(cls) -> None:
        """Load the configuration file(s).

        Note that this method is automatically loaded the first time .get()
        is called.

        :param filename: if not None load this configuration file instead of
            the default config files
        """
        for config_file in KNOWN_CONFIG_FILES:
            if os.path.isfile(config_file):
                cls.load_file(config_file)



from __future__ import annotations

import json
import mimetypes
import tempfile
from contextlib import closing

from e3.event import EventHandler, unique_id
from e3.fs import rm
from e3.os.process import Run
from e3.sys import python_script


class S3Handler(EventHandler):
    """Event handler that relies on AWS S3."""

    def __init__(self, event_s3_url, log_s3_url, sse="AES256", profile=None):
        self.event_s3_url = event_s3_url
        self.log_s3_url = log_s3_url
        self.aws_profile = profile
        self.sse = sse

    @classmethod
    def decode_config(cls, config_str):
        event_s3_url, log_s3_url, sse, aws_profile = config_str.split(",", 3)
        return {
            "event_s3_url": event_s3_url,
            "log_s3_url": log_s3_url,
            "sse": sse,
            "profile": aws_profile if aws_profile else None,
        }

    def encode_config(self):
        return "{},{},{},{}".format(
            self.event_s3_url,
            self.log_s3_url,
            self.sse,
            self.aws_profile if self.aws_profile is not None else "",
        )

    def send_event(self, event):
        def s3_cp(from_path, s3_url):
            cmd = ["s3", "cp", "--sse=%s" % self.sse]
            if self.aws_profile:
                cmd.append("--profile=%s" % self.aws_profile)
            cmd += [from_path, s3_url]

            s3 = Run(python_script("aws") + cmd, output=None)
            return s3.status == 0

        # Push attachments to s3 and keep track of their url.
        s3_attachs = {}
        for name, attach in list(event.get_attachments().items()):
            attach_path = attach[0]
            # Push the attachment
            s3_url = f"{self.log_s3_url}/{event.uid}/{name}"
            success = s3_cp(attach_path, s3_url)
            if not success:
                return False
            else:
                ctype, encoding = mimetypes.guess_type(attach_path)
                s3_attachs[name] = {
                    "s3_url": s3_url,
                    "encoding": encoding,
                    "ctype": ctype,
                }

        # Create the JSON to send on the event bucket
        s3_event = {"attachments": s3_attachs, "event": event.as_dict()}

        try:
            tempfile_name = None
            with closing(tempfile.NamedTemporaryFile(mode="w", delete=False)) as fd:
                tempfile_name = fd.name
                json.dump(s3_event, fd)

            # Note that an event can be sent several times with a different
            # status. As a consequence the target url in s3 should be different
            # for call to send.
            success = s3_cp(
                tempfile_name, f"{self.event_s3_url}/{event.uid}-{unique_id()}.s3",
            )

            if not success:
                return False
            else:
                return True
        finally:
            if tempfile_name is not None:
                rm(tempfile_name)



"""Support for reading and writing tar and zip archives."""


from __future__ import annotations

import fnmatch
import os
import sys
import tarfile
import tempfile
import zipfile
from contextlib import closing
from typing import TYPE_CHECKING

import e3
import e3.error
import e3.fs
import e3.log
import e3.os.fs


if TYPE_CHECKING:
    from typing import Callable, List, Optional, Sequence

logger = e3.log.getLogger("archive")


class E3ZipFile(zipfile.ZipFile):
    """Override default ZipFile with attributes preservation."""

    def _extract_member(self, member, path, pwd):
        result = super()._extract_member(member, path, pwd)

        if sys.platform != "win32":
            # Try to preserve attributes on non Windows platforms as
            # executable attribute is relevant on those platforms. As we rely
            # on an internal ignore any errors at this stage.
            try:
                # preserve bits 0-8 only: rwxrwxrwx
                # this come from a proposed patch on python.org
                # see: https://bugs.python.org/issue15795
                if not isinstance(member, zipfile.ZipInfo):
                    member = self.getinfo(member)
                attr = member.external_attr >> 16 & 0x1FF
                if attr != 0:
                    os.chmod(result, attr)
            except AttributeError:
                pass
        return result


class ArchiveError(e3.error.E3Error):
    pass


def is_known_archive_format(filename: str) -> bool:
    """Check if a given path is a supported archive format.

    :param filename: path
    :return: True if the path corresponding to a supported archive format
    """
    ext = e3.fs.extension(filename)
    return ext in (".tar.gz", ".tgz", ".tar.bz2", ".tar", ".zip")


def check_type(filename: str, force_extension: Optional[str] = None) -> str:
    """Return the archive extension.

    Internal function used by create_archive and unpack_archive.

    :param filename: the name of the archive to extract the extension
    :param force_extension: specify the archive extension if not in the
        filename

    :return: the file extension
    """
    # Check extension
    if (
        filename.endswith(".tar.gz")
        or filename.endswith(".tgz")
        or (force_extension is not None and force_extension in [".tar.gz", ".tgz"])
    ):
        ext = "tar.gz"
    elif filename.endswith(".tar.bz2") or (
        force_extension is not None and force_extension == ".tar.bz2"
    ):
        ext = "tar.bz2"
    elif filename.endswith(".tar") or (
        force_extension is not None and force_extension == ".tar"
    ):
        ext = "tar"
    elif filename.endswith(".zip") or (
        force_extension is not None and force_extension == ".zip"
    ):
        ext = "zip"
    else:
        raise ArchiveError(
            origin="unpack_archive", message='unknown format "%s"' % filename
        )
    return ext


def unpack_archive(
    filename: str,
    dest: str,
    selected_files: Optional[Sequence[str]] = None,
    remove_root_dir: bool = False,
    unpack_cmd: Optional[Callable[..., None]] = None,
    force_extension: Optional[str] = None,
    delete: bool = False,
    ignore: Optional[List[str]] = None,
    preserve_timestamps: bool = True,
) -> None:
    """Unpack an archive file (.tgz, .tar.gz, .tar or .zip).

    :param filename: archive to unpack
    :param dest: destination directory (should exist)
    :param selected_files: list of files to unpack (partial extraction). If
        None all files are unpacked
    :param remove_root_dir: if True then the root dir of the archive is
        suppressed.
        if set to 'auto' then the root dir of the archive is suppressed only
        if it is possible. If not do not raise an exception in that case and
        fallback on the other method.
    :param unpack_cmd: command to run to unpack the archive, if None use
        default methods or raise ArchiveError if archive format is not
        supported. If unpack_cmd is not None, then remove_root_dir is ignored.
        The unpack_cmd must raise ArchiveError in case of failure.
    :param force_extension: specify the archive extension if not in the
        filename. If filename has no extension and force_extension is None
        unpack_archive will fail.
    :param delete: if True and remove_root_dir is also True, remove files
        from dest if they do not exist in the archive
    :param ignore: a list of files/folders to keep when synchronizing with
        the final destination directory.
    :param preserve_timestamps: if False and remove_root_dir is True, and the
        target directory exists, ensure that updated files get their timestamp
        updated to current time.

    :raise ArchiveError: in case of error

    cygpath (win32) utilities might be needed when using remove_root_dir option
    """
    logger.debug("unpack %s in %s", filename, dest)
    # First do some checks such as archive existence or destination directory
    # existence.
    if not os.path.isfile(filename):
        raise ArchiveError(origin="unpack_archive", message="cannot find %s" % filename)

    if not os.path.isdir(dest):
        raise ArchiveError(
            origin="unpack_archive", message="dest dir %s does not exist" % dest
        )

    if selected_files is None:
        selected_files = []

    # We need to resolve to an absolute path as the extraction related
    # processes will be run in the destination directory
    filename = os.path.abspath(filename)

    if unpack_cmd is not None:
        # Use user defined unpack command
        if not selected_files:
            return unpack_cmd(filename, dest)
        else:
            return unpack_cmd(filename, dest, selected_files=selected_files)

    ext = check_type(filename, force_extension=force_extension)

    # If remove_root_dir is set then extract to a temp directory first.
    # Otherwise extract directly to the final destination
    if remove_root_dir:
        tmp_dest = tempfile.mkdtemp(
            prefix="", dir=os.path.dirname(os.path.abspath(dest))
        )
    else:
        tmp_dest = dest

    try:
        if ext in ("tar", "tar.bz2", "tar.gz"):
            try:
                # Set the right mode
                mode = "r:"
                if ext.endswith("bz2"):
                    mode += "bz2"
                elif ext.endswith("gz"):
                    mode += "gz"
                # Extract tar files
                with closing(tarfile.open(filename, mode=mode)) as fd:
                    check_selected = set(selected_files)

                    def is_match(name: str, files: Sequence[str]) -> bool:
                        """check if name match any of the expression in files.

                        :param name: file name
                        :param files: list of patterns to test against
                        :return: True when the name is matched
                        """
                        for pattern in files:
                            if fnmatch.fnmatch(name, pattern):
                                if pattern in check_selected:
                                    check_selected.remove(pattern)
                                return True
                        return False

                    dirs: List[str] = []

                    # IMPORTANT: don't use the method extract. Always use the
                    # extractall function. Indeed extractall will set file
                    # permissions only once all selected members are unpacked.
                    # Using extract can lead to permission denied for example
                    # if a read-only directory is created.
                    if selected_files:
                        member_list = []
                        for tinfo in fd:
                            if is_match(
                                tinfo.name, selected_files
                            ) or tinfo.name.startswith(tuple(dirs)):
                                # If dir then add it for recursive extracting
                                if tinfo.isdir() and not tinfo.name.startswith(
                                    tuple(dirs)
                                ):
                                    dirs.append(tinfo.name)
                                member_list.append(tinfo)

                        if check_selected:
                            raise ArchiveError(
                                "unpack_archive", "Cannot untar %s " % filename
                            )

                        fd.extractall(path=tmp_dest, members=member_list)
                    else:
                        fd.extractall(path=tmp_dest)

            except tarfile.TarError as e:
                raise ArchiveError(
                    origin="unpack_archive", message=f"Cannot untar {filename} ({e})",
                ).with_traceback(sys.exc_info()[2])

        else:
            try:
                with closing(E3ZipFile(filename, mode="r")) as zip_fd:
                    zip_fd.extractall(
                        tmp_dest, selected_files if selected_files else None
                    )
            except zipfile.BadZipfile as e:
                raise ArchiveError(
                    origin="unpack_archive", message=f"Cannot unzip {filename} ({e})",
                ).with_traceback(sys.exc_info()[2])

        if remove_root_dir:
            # First check that we have only one dir in our temp destination,
            # and no other files or directories. If not raise an error.
            nb_files = len(os.listdir(tmp_dest))
            if nb_files == 0:
                # Nothing to do...
                return
            if nb_files > 1:
                if remove_root_dir != "auto":
                    raise ArchiveError(
                        origin="unpack_archive",
                        message="archive does not have a unique root dir",
                    )

                # We cannot remove root dir but remove_root_dir is set to
                # 'auto' so fallback on non remove_root_dir method
                if not os.listdir(dest):
                    e3.fs.mv(os.path.join(tmp_dest, "*"), dest)
                else:
                    e3.fs.sync_tree(
                        tmp_dest,
                        dest,
                        delete=delete,
                        ignore=ignore,
                        preserve_timestamps=preserve_timestamps,
                    )
            else:
                root_dir = os.path.join(tmp_dest, os.listdir(tmp_dest)[0])

                # Now check if the destination directory is empty. If this is
                # the case a simple move will work, otherwise we need to do a
                # sync_tree (which cost more)

                if not os.listdir(dest):
                    e3.fs.mv(
                        [os.path.join(root_dir, f) for f in os.listdir(root_dir)], dest
                    )
                else:
                    e3.fs.sync_tree(
                        root_dir,
                        dest,
                        delete=delete,
                        ignore=ignore,
                        preserve_timestamps=preserve_timestamps,
                    )

    finally:
        # Always remove the temp directory before exiting
        if remove_root_dir:
            e3.fs.rm(tmp_dest, True)


def create_archive(
    filename: str,
    from_dir: str,
    dest: str,
    force_extension: Optional[str] = None,
    from_dir_rename: Optional[str] = None,
    no_root_dir: bool = False,
) -> None:
    """Create an archive file (.tgz, .tar.gz, .tar or .zip).

    On Windows, if the python tarfile and zipfile modules are available, the
    python implementation is used to create the archive.  On others system,
    create_archive spawn tar, gzip or zip as it is twice faster that the python
    implementation. If the tar, gzip or zip binary is not found, the python
    implementation is used.

    :param filename: archive to create
    :param from_dir: directory to pack (full path)
    :param dest: destination directory (should exist)
    :param force_extension: specify the archive extension if not in the
        filename. If filename has no extension and force_extension is None
        create_archive will fail.
    :param from_dir_rename: name of root directory in the archive.
    :param no_root_dir: create archive without the root dir (zip only)

    :raise ArchiveError: if an error occurs
    """
    # Check extension
    from_dir = from_dir.rstrip("/")
    filepath = os.path.abspath(os.path.join(dest, filename))

    ext = check_type(filename, force_extension=force_extension)

    if from_dir_rename is None:
        from_dir_rename = os.path.basename(from_dir)

    if ext == "zip":
        zip_archive = zipfile.ZipFile(filepath, "w", zipfile.ZIP_DEFLATED)
        for root, _, files in os.walk(from_dir):
            relative_root = os.path.relpath(
                os.path.abspath(root), os.path.abspath(from_dir)
            )
            for f in files:
                zip_file_path = os.path.join(from_dir_rename, relative_root, f)
                if no_root_dir:
                    zip_file_path = os.path.join(relative_root, f)
                zip_archive.write(os.path.join(root, f), zip_file_path)
        zip_archive.close()
    else:
        if ext == "tar":
            tar_format = "w"
        elif ext == "tar.gz":
            tar_format = "w:gz"
        elif ext == "tar.bz2":
            tar_format = "w:bz2"
        else:  # defensive code
            raise ArchiveError(f"unsupported format {tar_format}")
        with closing(tarfile.open(filepath, tar_format)) as tar_archive:
            tar_archive.add(name=from_dir, arcname=from_dir_rename, recursive=True)



from __future__ import annotations

import collections
from typing import TYPE_CHECKING

import e3.os
import e3.os.platform
from e3.platform_db import get_knowledge_base

if TYPE_CHECKING:
    from typing import Any, Dict, Optional

KNOWLEDGE_BASE = get_knowledge_base()


# noinspection PyUnresolvedReferences
class Platform(
    collections.namedtuple(
        "Platform",
        [
            "cpu",
            "os",
            "is_hie",
            "platform",
            "triplet",
            "machine",
            "domain",
            "is_host",
            "is_default",
        ],
    )
):
    """Class that allow user to retrieve os/cpu specific information.

    Attributes are:

    - cpu: CPU information
    - os: Operating System information
    - is_hie: whether the system is a high integrity system
    - platform: the platform name, e.g. arm-elf-linux
    - triplet: the GCC target
    - machine: machine name
    - domain: domain name
    - is_host: True if the instance represent information for the current
        machine
    - is_default: True if the platform is the default one
    """

    default_arch: Optional[Platform] = None
    system_info = e3.os.platform.SystemInfo

    __slots__ = ()

    @classmethod
    def get(
        cls,
        platform_name: Optional[str] = None,
        version: Optional[str] = None,
        machine: Optional[str] = None,
        mode: Optional[str] = None,
        compute_default: bool = False,
    ) -> Platform:
        """Return a Platform object.

        :param platform_name: if None or empty then automatically detect
            current platform (native). Otherwise should be a valid platform
            string.
        :param version:  if None, assume default OS version or find it
            automatically (native case only). Otherwise should be a valid
            version string.
        :param machine: name of the machine
        :param compute_default: if True compute the default Arch for the
            current machine (this parameter is for internal purpose only).
        :param mode: an os mode (ex: rtp for vxworks)
        """
        # normalize arguments
        if not version:
            version = e3.os.platform.UNKNOWN
        if machine is None or machine == e3.os.platform.UNKNOWN:
            machine = ""
        if not mode:
            mode = e3.os.platform.UNKNOWN

        # Initialize default arch class variable
        if cls.default_arch is None and not compute_default:
            cls.default_arch = Platform.get(compute_default=True)

        is_default = False
        is_host = False
        domain = e3.os.platform.UNKNOWN

        if compute_default:
            default_platform = cls.system_info.platform()
        else:
            assert cls.default_arch is not None
            default_platform = cls.default_arch.platform

        # Check if the object correspond to the current machine and thus allow
        # us to compute some additional info automatically
        if platform_name in (None, "", "default"):
            platform_name = default_platform
            is_host = True
        if machine == cls.system_info.hostname()[0]:
            is_host = True

        if is_host:
            # This is host so we can guess the machine name and domain
            machine, domain = cls.system_info.hostname()
            is_default = platform_name == default_platform

        # Fill other attributes
        pi = KNOWLEDGE_BASE.platform_info[platform_name]
        cpu = e3.os.platform.CPU.get(pi["cpu"], pi.get("endian", None), is_host)
        os = e3.os.platform.OS.get(pi["os"], is_host, version=version, mode=mode)
        is_hie = pi["is_hie"]

        # Find triplet
        triplet = KNOWLEDGE_BASE.build_targets[platform_name]["name"] % {
            "os_version": os.version
        }

        return cls(
            cpu,
            os,
            is_hie,
            platform_name,
            triplet,
            machine,
            domain,
            is_host,
            is_default,
        )

    @property
    def is_virtual(self) -> bool:
        """Check if we are on a virtual system.

        :return: True if the system represented by Arch is a virtual machine
        """
        if not self.is_host:
            return False
        return self.system_info.is_virtual()

    def to_dict(self) -> Dict[str, Any]:
        """Export os and cpu variables as os_{var} and cpu_{var}.

        :return: a dictionary representing the current Arch instance
        """
        str_dict = self._asdict()
        str_dict["is_virtual"] = self.is_virtual

        for key, var in self.os.as_dict().items():
            str_dict["os_" + key] = var
        for key, var in self.cpu.as_dict().items():
            str_dict["cpu_" + key] = var
        del str_dict["os"]
        del str_dict["cpu"]
        return str_dict

    def __str__(self) -> str:
        """Return a representation string of the object."""
        result = (
            "platform: %(platform)s\n"
            "machine:  %(machine)s\n"
            "is_hie:   %(is_hie)s\n"
            "is_host:  %(is_host)s\n"
            "is_virtual: %(is_virtual)s\n"
            "triplet:  %(triplet)s\n"
            "domain:   %(domain)s\n"
            "OS\n"
            "   name:          %(os_name)s\n"
            "   version:       %(os_version)s\n"
            "   exeext:        %(os_exeext)s\n"
            "   dllext:        %(os_dllext)s\n"
            "   is_bareboard:  %(os_is_bareboard)s\n"
            "CPU\n"
            "   name:   %(cpu_name)s\n"
            "   bits:   %(cpu_bits)s\n"
            "   endian: %(cpu_endian)s\n"
            "   cores:  %(cpu_cores)s" % self.to_dict()
        )
        return result



from __future__ import annotations

import json
import os

from e3.event import EventHandler
from e3.fs import cp, mkdir


class FileHandler(EventHandler):
    def __init__(self, log_dir):
        self.log_dir = log_dir

    def send_event(self, event):
        d = event.as_dict()
        prefix = "{}-{}".format(d["name"], d["uid"])
        log_file = os.path.join(self.log_dir, prefix + ".log")
        attach_dir = os.path.join(self.log_dir, prefix)
        mkdir(attach_dir)
        with open(log_file, "w") as fd:
            json.dump(d, fd, indent=2, sort_keys=True)
        for name, attachment in list(event.get_attachments().items()):
            cp(attachment[0], os.path.join(attach_dir, name))
        return True

    @classmethod
    def decode_config(cls, config_str):
        return {"log_dir": config_str}

    def encode_config(self):
        return self.log_dir



from __future__ import annotations

from e3.anod.error import AnodError
from e3.main import Main


def anod_cmdline(subparsers, action_name, action_help):
    parser = subparsers.add_parser(action_name, help=action_help)
    parser.set_defaults(action_name=action_name)
    parser.add_argument("spec", metavar="SPEC", help="The specification file name")
    parser.add_argument("--qualifier")


def anod():
    """bin/anod script entry point.

    This script is installed in the sandbox.
    """
    import os
    import sys

    import e3.anod.driver
    import e3.anod.loader
    import e3.anod.sandbox
    import e3.env
    import e3.store
    import e3.store.cache

    sandbox_dir = os.path.abspath(
        os.path.join(os.path.dirname(sys.modules["__main__"].__file__), os.pardir)
    )

    sandbox = e3.anod.sandbox.SandBox()
    sandbox.root_dir = sandbox_dir

    # Load the local specs
    spec_repo = e3.anod.loader.AnodSpecRepository(sandbox.specs_dir)

    # Load the cache
    cache = e3.store.cache.load_cache(
        "file-cache", {"cache_dir": sandbox.tmp_cache_dir}
    )

    store = e3.store.load_store("http-simple-store", {}, cache)

    m = Main()
    subparsers = m.argument_parser.add_subparsers()
    anod_cmdline(subparsers, "download", "download a binary package")
    m.parse_args()

    action = m.args.action_name
    spec = m.args.spec
    qualifier = m.args.qualifier

    anod_cls = spec_repo.load(name=spec)
    anod_instance = anod_cls(
        qualifier=qualifier, kind=action, jobs=1, env=e3.env.BaseEnv.from_env()
    )

    # ??? inject the sandbox
    anod_instance.sandbox = sandbox

    driver = e3.anod.driver.AnodDriver(anod_instance=anod_instance, store=store)

    try:
        driver.activate(sandbox, spec_repo)
        driver.call(action)
    except AnodError as err:
        print(err, file=sys.stderr)
        sys.exit(1)



"""Modification of the yaml loader for E3."""


from __future__ import annotations

import os
import re
import sys
from collections import OrderedDict
from typing import TYPE_CHECKING

import yaml
import yaml.constructor
import yaml.parser

import e3.log
from e3.text import format_with_dict

if TYPE_CHECKING:
    # Conditonal imports do not work with mypy, unconditionaly use yaml.Loader
    # for type checking
    from typing import Any, IO, List, Set, Tuple, Union
    from yaml import Loader
else:
    try:
        from yaml import CLoader as Loader
    except ImportError:
        from yaml import Loader


class YamlError(Exception):
    pass


class OrderedDictYAMLLoader(Loader):
    """A YAML loader that loads mappings into ordered dictionaries.

    The loader also support the !include constructor that allows
    inclusion of yaml files into another yaml
    """

    def __init__(self, stream: IO[str]):
        self.name = None
        self.stream = stream
        super().__init__(stream)

        self.add_constructor(
            "tag:yaml.org,2002:map", type(self).construct_yaml_map
        )  # type: ignore
        self.add_constructor(
            "tag:yaml.org,2002:omap", type(self).construct_yaml_map
        )  # type: ignore
        self.add_constructor("!include", type(self).yaml_include)  # type: ignore

    def yaml_include(self, node):
        # Get the path out of the yaml file
        if self.name is None:
            if not isinstance(self.stream, str) or isinstance(self.stream, str):
                self.name = getattr(self.stream, "name", None)

        if self.name is not None and os.path.isfile(self.name):
            file_name = os.path.join(os.path.dirname(self.name), node.value)
        else:
            file_name = node.value

        with open(file_name, "rb") as inputfile:
            return yaml.load(inputfile, OrderedDictYAMLLoader)

    def construct_yaml_map(self, node):
        data = OrderedDict()
        yield data
        value = self.construct_mapping(node)
        data.update(value)

    def construct_mapping(self, node, deep=False):
        if isinstance(node, yaml.MappingNode):
            self.flatten_mapping(node)
        else:
            raise yaml.constructor.ConstructorError(
                context=None,
                context_mark=None,
                problem="expected a mapping node, but found %s" % node.id,
                problem_mark=node.start_mark,
            )

        mapping = OrderedDict()
        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            try:
                hash(key)
            except TypeError as exc:
                raise yaml.constructor.ConstructorError(
                    context="while constructing a mapping",
                    context_mark=node.start_mark,
                    problem="found unacceptable key (%s)" % exc,
                    problem_mark=key_node.start_mark,
                )
            value = self.construct_object(value_node, deep=deep)
            if key in mapping:
                raise yaml.constructor.ConstructorError(
                    context="while constructing a mapping",
                    context_mark=node.start_mark,
                    problem="found duplicate key (%s)" % key,
                    problem_mark=key_node.start_mark,
                )
            mapping[key] = value
        return mapping


def load_ordered(filename: str) -> OrderedDict:
    """Load a .yaml file, keep the file order."""
    with open(filename) as f:
        return yaml.load(f, OrderedDictYAMLLoader)


class CaseParser:
    """Parse case statements in an OrderedDict.

    Each time a key starting with ``case_`` (or the prefix you choose) if
    found, in a block mapping, the value of the sub block matching the key
    value is added to the result dictionary.

    For instance: with an initial config
    ``{'param1': 'full', 'param2': 'short'}``, the result of the parsing of:

    .. code-block:: yaml

        {'case_param1': {
            'full': {
                'case_param2': {
                    'full': {'a': 2, 'b': 1, 'c': 'content'},
                    'short': {'y': 0, 'c': ''}}},
            'short': {
                'case_param2': {
                    'full': {'a': 9, 'b': 5, 'c': 'default'},
                    'short': {'y': 3, 'c': ''}}}},
         'value1': 10,
         'value2': '%(c)s',
         'case_param2' : {
                'f.*l': {'value3': 30, 'a': 42}}}

    is ``{'y': 0, 'c': '', 'value2': '', 'value1': 10}``

    and with ``{'param1': 'short', 'param2': 'full'}`` we get:

    .. code-block:: yaml

          {'a': 42, 'c': 'default', 'b': 5, 'value3': 30,
           'value2': 'default', 'value1': 10}

    Note that values can be redefined and that you can add a python regexp, as
    supported by the `re` module, in your case values.
    """

    def __init__(self, initial_config: dict, case_prefix: str = "case_"):
        self.__state = initial_config.copy()
        self.case_prefix = case_prefix

        # This contains the list of keys that have been updated. This
        # allow us to remove the keys part of initial_config that are
        # not modified.
        self.keys: Set[str] = set()

    def __parse_case(self, case_key: str, data: dict) -> Any:
        """Parse a case statement.

        :param case_key: the variable on which the case is evaluated
        :param data: the dictionary of case conditions

        :return: the value of the matched element or None
        """
        key = case_key[len(self.case_prefix) :]
        key_val = str(self.__state[key])

        result = next(
            ((key_val, k, data[k]) for k in data if re.match("^%s$" % k, key_val)), None
        )
        if result is not None:
            e3.log.debug("%s=%s match %s", key, result[0], result[1])
            return result[2]
        else:
            return None

    def __format_value(self, value: Any) -> Any:
        """Format a value.

        :param value: the value to be formatted

        :return: the result of the expansion
        """
        try:
            if isinstance(value, str):
                return format_with_dict(value, self.__state)
            elif isinstance(value, dict):
                return {k: self.__format_value(v) for k, v in value.items()}
            elif isinstance(value, list):
                return [self.__format_value(d) for d in value]
        except (KeyError, TypeError):
            e3.log.debug("Cannot format %s, ignore it", value)

        return value

    def __update_state(self, key: str, value: Any, cursor: Any, prefix: Tuple) -> None:
        """Update state.

        :param key: the key to modify. Leading or trailing '+' in the key name
            are interpreted respectively as append and prepend operators.
            For dictionaries these operators are interpreted as an update
            on the original value.
        :param value: the new value
        :param cursor: the object to be updated
        :param prefix: a tuple of string that gives the position of cursor in
            self.__state. This is used only for debugging purposes
        """
        real_key = key.strip("+")
        real_value = self.__format_value(value)

        # Update the list of keys that should be considered in the final
        # result
        if cursor is self.__state:
            self.keys.add(real_key)

        real_key_str = "[%s]" % "][".join(prefix + (real_key,))

        if real_key not in cursor or real_key == key:
            e3.log.debug("set %s -> %s", real_key_str, real_value)
            cursor[real_key] = real_value
        else:
            if isinstance(cursor[real_key], dict):
                e3.log.debug("update %s -> %s", real_key_str, real_value)
                cursor[real_key].update(real_value)
            else:
                if key.startswith("+"):
                    e3.log.debug("append %s -> %s", real_key_str, real_value)
                    cursor[real_key] = cursor[real_key] + real_value
                elif key.endswith("+"):
                    e3.log.debug("prepend %s -> %s", real_key_str, real_value)
                    cursor[real_key] = real_value + cursor[real_key]

    def parse(self, data: Any) -> Any:
        """Parse.

        :param data: a python object. Note that dictionaries in that structure
            should be OrderedDict.

        :return: a new python object after expansion of case statements and
            formatting of values
        """
        return self.__parse(data, self.__state, ())

    def __parse(self, data: Any, cursor: Any, prefix: tuple) -> Any:
        """Parse (internal).

        :param data: a python object. Note that dictionaries in that structure
            should be OrderedDict.
        :param cursor: a ref to a substructure of self.__state
        :param prefix: the current location in self.__state (a tuple of keys)

        :return: the new cursor object
        """
        if not isinstance(data, dict):
            return self.__format_value(data)

        for key in data:
            if key.startswith(self.case_prefix):
                pc = self.__parse_case(key, data[key])
                if pc is not None:
                    result = self.__parse(pc, cursor, prefix)
                    if not isinstance(result, dict):
                        assert len(list(data.keys())) == 1, "invalid configuration file"
                        return result
            else:
                subcursor = cursor.get(key.strip("+"), {})
                subprefix = prefix + (key.strip("+"),)
                self.__update_state(
                    key,
                    self.__parse(data[key], cursor=subcursor, prefix=subprefix),
                    cursor,
                    prefix,
                )

        if cursor is self.__state:
            return {k: v for k, v in cursor.items() if k in self.keys}
        else:
            return cursor


def load_with_config(filename: Union[str, List[str]], config: dict) -> Any:
    """Load yaml config files with case statement handling.

    :param filename: a path or list of path. When a list of path
        is given, config files are loaded in order each one
        updating the result of the previous parsing.
    :param config: initial state

    :return: the final object
    """
    if isinstance(filename, str):
        filename = [filename]

    result = None
    parser = CaseParser(config)

    for f in filename:
        try:
            e3.log.debug("load config file: %s", f)
            conf_data = load_ordered(f)
            result = parser.parse(conf_data)
        except OSError:
            raise YamlError("cannot read: %s" % f, "load_with_config").with_traceback(
                sys.exc_info()[2]
            )
        except (yaml.parser.ParserError, yaml.constructor.ConstructorError) as e:
            raise YamlError(
                f"{f} is an invalid yaml file: {e}", "load_with_config"
            ).with_traceback(sys.exc_info()[2])

    return result


def load_with_regexp_table(filename: str, selectors: List[str], data: dict) -> dict:
    """Load a yaml file using regexp tables.

    :param filename: the yaml file to load
    :param selectors: a list of string that will be used to match the regexps
        in table
    :param data: a dictionary used to replace part of value content

    This function expect a yaml file that has the following format::

        key1:
            [['regexp1_1', ..., 'regexp1_n', value],
             ['regexp2_1', ..., 'regexp2_n', value],...

        key2:
            ...

    The returned dictionary will have key1, key2, ... as keys. For each key
    the value is the last element of the first sublist for which the n first
    first regexp do match, strings in selectors list. So in the yaml file each
    sublist associated with each key should have exactly ``len(selectors) + 1``
    elements.
    """
    e3.log.debug("load %s with %s", filename, selectors)
    with open(filename) as f:
        conf_data = yaml.load(f.read(), OrderedDictYAMLLoader)

    assert isinstance(conf_data, dict), (
        "top level object in %s should be a dict" % filename
    )

    result = {}

    for key in conf_data:
        key_data = conf_data[key]
        assert isinstance(key_data, list), "value for key %s is not a list" % key

        for line in key_data:
            assert isinstance(line, list), (
                "value for key %s should be a list of list" % key
            )
            assert len(line) == len(selectors) + 1

            has_matched = True
            for index, r in enumerate(line[0:-1]):
                if len(r) == 0:
                    r = ".*"
                if not re.search(r"^%s$" % r, str(selectors[index])):
                    has_matched = False

            if has_matched:
                result[key] = line[-1]
                break

    # At this stage we have our dictionary filled. Now use data to replace to
    # replace %()s strings.

    for key in result:
        if isinstance(result[key], str):
            result[key] = result[key] % data
        elif isinstance(result[key], list):
            result[key] = [k % data for k in result[key]]

    e3.log.debug("yaml results: %s", result)
    return result



from __future__ import annotations

import base64
import json
from e3.date import utc_timestamp


def get_payload(token: str) -> dict:
    """Get payload from a JSON Web Token.

    :param token: token
    :return: decoded payload
    """
    data = {}
    if token.count(".") == 2:
        # Extract the payload
        signing_part, _ = token.rsplit(".", 1)
        _, payload = signing_part.split(".", 1)
        # Add required padding
        rem = len(payload) % 4
        if rem > 0:
            payload += "=" * (4 - rem)
        try:
            data = json.loads(base64.urlsafe_b64decode(payload).decode("utf-8"))
        except (ValueError, TypeError):
            pass
    return data


def is_valid(token: str) -> bool:
    """Return true if the token is still valid, false otherwise.

    :param token: full token
    :return: return true if the token is still valid, false otherwise
    """
    payload = get_payload(token)

    # Given that we will use the token if valid, keep some margin and
    # do not consider a token valid if it will be valid less than 5 min
    deadline = utc_timestamp() + 5 * 60

    return payload.get("typ") == "Bearer" and payload.get("exp", 0) > deadline



"""Low-level file manipulation.

All function here should be platform indepenent, should not involve globbing or
logging (unless in case of unexpected failure).
"""


from __future__ import annotations

import collections
import itertools
import os
import re
import shutil
import stat
import sys
from typing import TYPE_CHECKING

import e3
import e3.error
import e3.log

if TYPE_CHECKING:
    from typing import Any, Callable, List, Optional, Tuple, Union


class OSFSError(e3.error.E3Error):
    pass


logger = e3.log.getLogger("os_fs")


def cd(path: str) -> None:
    """Change current directory.

    :param path: directory name
    :raise OSFSError: in case of error
    """
    try:
        os.chdir(path)
    except Exception as e:
        logger.error(e, exc_info=True)
        raise OSFSError(
            origin="cd", message="can't chdir to %s\n" % path
        ).with_traceback(sys.exc_info()[2])


def chmod(mode: str, filename: str) -> int:
    """Chmod with interface similar to Unix tool.

    :param mode: should conform with posix specification for
        chmod utility (ex: +wx). See chmod man page for more information
    :param filename: the target file
    :return: the mode that has been set
    """
    # Developer note: for local variable names in this function
    # we try to use the words used in the opengroup specification
    # this way we can map easily between the implementation and
    # what is defined in the standard.

    whos = {"u": stat.S_IRWXU, "g": stat.S_IRWXG, "o": stat.S_IRWXO}
    perms = {"r": stat.S_IROTH, "w": stat.S_IWOTH, "x": stat.S_IXOTH}

    current_mode = os.stat(filename).st_mode

    # Retrieve umask
    umask = os.umask(0)
    os.umask(umask)

    clauses = mode.split(",")

    for clause in clauses:
        match = re.search(r"([ugoa]+)([-\+=].*)", clause)
        if match is not None:
            wholist = match.group(1)
            actionlist = match.group(2)
        else:
            wholist = ""
            actionlist = clause

        actions = re.findall(r"(?:([-\+=])?([ugo]|[0-7]+|[rwx]*))", actionlist)
        assert "".join(list(itertools.chain.from_iterable(actions))) == actionlist

        for (op, permlist) in actions:
            if permlist == "" and op != "=":
                continue
            else:
                if permlist in ("u", "g", "o"):
                    action_mask = current_mode & whos[permlist]
                    if permlist == "u":
                        action_mask >>= 6
                    elif permlist == "g":
                        action_mask >>= 3
                elif permlist.isdigit():
                    raise OSFSError(
                        origin="chmod",
                        message="numeric mode not supported," " use os.chmod instead",
                    )
                else:
                    action_mask = 0
                    for perm in permlist:
                        action_mask |= perms[perm]

                if wholist == "":
                    action_mask = action_mask | action_mask << 3 | action_mask << 6
                    action_mask &= ~umask
                    apply_mask = stat.S_IRWXO | stat.S_IRWXU | stat.S_IRWXG
                else:
                    if "a" in wholist:
                        action_mask = action_mask | action_mask << 3 | action_mask << 6
                        apply_mask = stat.S_IRWXO | stat.S_IRWXU | stat.S_IRWXG
                    else:
                        final_action_mask = 0
                        apply_mask = 0
                        for who in wholist:
                            if who == "u":
                                final_action_mask |= action_mask << 6
                                apply_mask |= stat.S_IRWXU
                            elif who == "g":
                                final_action_mask |= action_mask << 3
                                apply_mask |= stat.S_IRWXG
                            else:
                                final_action_mask |= action_mask
                                apply_mask |= stat.S_IRWXO

                        action_mask = final_action_mask
                if op == "-":
                    current_mode &= ~action_mask
                elif op == "=":
                    current_mode = (current_mode & ~apply_mask) | action_mask
                else:
                    current_mode = current_mode | action_mask

    os.chmod(filename, current_mode)
    return current_mode


def df(path: str, full: bool = False) -> Union[int, Tuple]:
    """Disk space available on the filesystem containing the given path.

    :param path: a path
    :param full: if True return full disk information otherwise only
        space left.

    :return: either space left in Mo or a :py:func:`collections.namedtuple`
        with ``total``, ``used`` and ``free`` attributes. Each attribute is
        an int representing Mo.
    """
    _ntuple_diskusage = collections.namedtuple("usage", "total used free")
    if sys.platform == "win32":  # unix: no cover
        import ctypes

        path = ctypes.c_wchar_p(path)
        GetDiskFreeSpaceEx = ctypes.WINFUNCTYPE(
            ctypes.c_int,
            ctypes.c_wchar_p,
            ctypes.POINTER(ctypes.c_uint64),
            ctypes.POINTER(ctypes.c_uint64),
            ctypes.POINTER(ctypes.c_uint64),
        )
        GetDiskFreeSpaceEx = GetDiskFreeSpaceEx(
            ("GetDiskFreeSpaceExW", ctypes.windll.kernel32),
            ((1, "path"), (2, "freeuserspace"), (2, "totalspace"), (2, "freespace")),
        )

        def GetDiskFreeSpaceEx_errcheck(result, func, args):
            del func
            if not result:  # defensive code
                raise ctypes.WinError()
            return (args[1].value, args[2].value, args[3].value)

        GetDiskFreeSpaceEx.errcheck = GetDiskFreeSpaceEx_errcheck
        _, total, free = GetDiskFreeSpaceEx(path)
        used = total - free
    else:  # windows: no cover
        # f_frsize = fundamental filesystem block size
        # f_bsize = preferred file system block size
        # The use of f_frsize seems to give more accurate results.
        st = os.statvfs(path)
        free = st.f_bavail * st.f_frsize
        total = st.f_blocks * st.f_frsize
        used = (st.f_blocks - st.f_bfree) * st.f_frsize
    if full:
        return _ntuple_diskusage(
            total // (1024 * 1024), used // (1024 * 1024), free // (1024 * 1024)
        )
    return free // (1024 * 1024)


def __safe_unlink_func() -> Tuple[Callable[[str], None], Callable[[str], None]]:
    """Provide a safe unlink function on windows.

    Note that all this is done to ensure that rm is working fine on Windows 7
    and 2008R2. Indeed very often, deletion will fail with access denied
    error. The typical scenario is when you spawn an executable and try to
    delete it just afterward.
    """
    if sys.platform == "win32":  # unix: no cover
        from e3.os.windows.fs import NTFile

        def win_rm(x: str) -> None:
            return NTFile(x).unlink()

        return win_rm, win_rm
    else:  # windows: no cover

        def os_remove(x: str) -> None:
            return os.remove(x)

        def os_rmdir(x: str) -> None:
            return os.rmdir(x)

        return os_remove, os_rmdir


safe_remove, safe_rmdir = __safe_unlink_func()


def force_remove_file(path: str) -> None:
    """Force file removing, changing permissions if first attempt failed.

    :param path: path of the file to remove
    """
    try:
        safe_remove(path)
    except OSError:
        # The permission of the parent directory does not allow us to remove
        # the file, temporary get write permission in the directory
        dir_path = os.path.dirname(path)
        orig_mode = os.stat(dir_path).st_mode
        chmod("u+w", dir_path)

        # ??? It seems that this might be needed on windows
        os.chmod(path, 0o700)
        safe_remove(path)
        os.chmod(dir_path, orig_mode)


def max_path() -> int:
    """Return the maximum length for a path.

    :return: the maximum length
    """
    if sys.platform == "win32":  # unix: no cover
        from ctypes.wintypes import MAX_PATH

        return MAX_PATH
    else:  # windows: no cover
        return os.pathconf("/", "PC_PATH_MAX")


def mv(source: str, target: str) -> None:
    """Move a file.

    :param target: file to move
    :param source: target file or directory
    """
    # Compute file list and number of file to copy
    if os.path.isdir(source) and os.path.isdir(target):
        shutil.move(source, os.path.join(target, os.path.basename(source)))
    else:
        shutil.move(source, target)


def touch(filename: str) -> None:
    """Update file access and modification times. Create the file if needed.

    :param filename: file to update
    """
    if os.path.exists(filename):
        os.utime(filename, None)
    else:
        with open(filename, "w+"):
            pass


def unixpath(path: str) -> str:
    r"""Convert path to Unix/Cygwin format.

    :param path: path string to convert
    :return: the converted path

    On Unix systems this function is identity. On Win32 systems it removes
    drive letter information and replace \\ by /.
    """
    if path and sys.platform == "win32":  # unix: no cover
        # Cygpath is not available so just replace \ by / and remove drive
        # information. This should work in most cases
        result = path.replace("\\", "/")
        m = re.match("[a-zA-Z]:(.*)", result)
        if m is not None:
            result = m.group(1)
        return result
    else:  # windows: no cover
        return path


def which(prog: str, paths: Optional[str] = None, default: Any = "") -> Any:
    """Locate executable.

    :param prog: program to find
    :param paths: if not None then we use this value instead of PATH to look
        for the executable.
    :param default: default value to return if not found

    :return: absolute path to the program on success, found by searching for an
      executable in the directories listed in the environment variable PATH
      or default value if not found
    """

    def is_exe(file_path: str) -> bool:
        return os.path.isfile(file_path) and os.access(file_path, os.X_OK)

    def possible_names(file_path: str) -> List[str]:
        names = [file_path]
        if sys.platform == "win32":  # unix: no cover
            names.extend(
                [file_path + ext for ext in os.environ.get("PATHEXT", "").split(";")]
            )
        return names

    fpath, _ = os.path.split(prog)
    if fpath:
        # Full path given, check if executable
        for progname in possible_names(prog):
            if is_exe(progname):
                return progname
    else:
        # Check for all directories listed in $PATH
        if paths is None:
            paths = os.environ["PATH"]

        for pathdir in paths.split(os.pathsep):
            exe_file = os.path.join(pathdir, prog)
            for progname in possible_names(exe_file):
                if is_exe(progname):
                    return progname

    # Not found.
    return default



import os
import subprocess

from e3.fs import echo_to_file, rm
from e3.os.fs import unixpath
from e3.vcs.git import GitError, GitRepository

import pytest
import tempfile
from contextlib import closing


@pytest.mark.git
def test_git_non_utf8():
    """Test with non utf-8 encoding in changelog."""
    working_tree = os.path.join(os.getcwd(), "working_tree")
    repo = GitRepository(working_tree)
    repo.init()
    os.chdir(working_tree)
    new_file = os.path.join(working_tree, "new.txt")
    commit_msg = os.path.join(working_tree, "commit.txt")

    with open(commit_msg, "wb") as fd:
        fd.write(b"\x03\xff")

    with open(new_file, "wb") as fd:
        fd.write(b"\x03\xff")

    repo.git_cmd(["add", "new.txt"])
    repo.git_cmd(["config", "user.email", "e3-core@example.net"])
    repo.git_cmd(["config", "user.name", "e3 core"])
    repo.git_cmd(["commit", "-F", commit_msg])

    with closing(tempfile.NamedTemporaryFile(mode="w", delete=False)) as fd:
        repo.write_log(fd)
        tmp_filename = fd.name
    try:
        with open(tmp_filename) as fd:
            commits = list(repo.parse_log(fd, max_diff_size=1024))
    finally:
        rm(tmp_filename)

    assert "\\x03\\xff" in commits[0]["diff"]


@pytest.mark.git
def test_git_repo():
    working_tree = os.path.join(os.getcwd(), "working_tree")
    working_tree2 = os.path.join(os.getcwd(), "working_tree2")
    repo = GitRepository(working_tree)
    repo.init()
    os.chdir(working_tree)
    new_file = os.path.join(working_tree, "new.txt")
    echo_to_file(new_file, "new\n")

    commit_note = unixpath(os.path.join(working_tree, "commit_note.txt"))
    echo_to_file(
        commit_note,
        "\n".join(
            (
                "Code-Review+2: Nobody <nobody@example.com>",
                "Submitted-by: Nobody <nobody@example.com>",
                "Submitted-at: Thu, 08 Jun 2017 18:40:11 +0200",
            )
        ),
    )

    repo.git_cmd(["add", "new.txt"])
    repo.git_cmd(["config", "user.email", "e3-core@example.net"])
    repo.git_cmd(["config", "user.name", "e3 core"])
    repo.git_cmd(["commit", "-m", "new file"])
    repo.git_cmd(["tag", "-a", "-m", "new tag", "20.0855369232"])
    repo.git_cmd(["notes", "--ref", "review", "add", "HEAD", "-F", commit_note])

    # try with gerrit notes
    with open("log.txt", "w") as f:
        repo.write_log(f, with_gerrit_notes=True)
    with open("log.txt") as f:
        commits = list(repo.parse_log(f))
        assert "nobody@example.com" in commits[0]["notes"]["Code-Review+2"]

    # try with an invalid note
    repo.git_cmd(
        ["notes", "--ref", "review", "add", "HEAD", "-f", "-m", "invalid-note"]
    )
    with open("log.txt", "w") as f:
        repo.write_log(f, with_gerrit_notes=True)
    with open("log.txt") as f:
        commits = list(repo.parse_log(f))
        assert commits[0]["notes"] is None

    # try again without gerrit notes
    with open("log.txt", "w") as f:
        repo.write_log(f)
    with open("log.txt") as f:
        commits = list(repo.parse_log(f))
        assert "new file" in commits[0]["message"]
        assert commits[0]["email"] == "e3-core@example.net"
        new_sha = commits[0]["sha"]

    assert "20.0855369232" in repo.describe()
    assert new_sha == repo.rev_parse()

    with pytest.raises(GitError) as err:
        repo.describe("g")
    assert "describe --always g" in str(err)

    echo_to_file(new_file, "new line\n", append=True)

    with open("commit1.diff", "wb") as f:
        repo.write_local_diff(f)

    with open("commit1.diff", "rb") as f:
        assert b"+new line" in f.read()

    echo_to_file(new_file, 10000 * "*")

    repo.git_cmd(["commit", "-a", "-m", "file update"])
    with open("log2.txt", "w") as f:
        repo.write_log(f)
    with open("log2.txt") as f:
        commits = list(repo.parse_log(f, max_diff_size=1000))
        # assert b'diff too long' not in commits[1]['diff']
        assert "file update" in commits[0]["message"]
        assert "diff too long" in commits[0]["diff"]
        assert "new file" in commits[1]["message"]
        assert commits[1]["sha"] != commits[0]["sha"]
        assert commits[1]["diff"] != commits[0]["diff"]

    repo2 = GitRepository(working_tree2)
    giturl = "file://%s" % working_tree.replace("\\", "/")
    repo2.init(url=giturl, remote="tree1")
    repo2.update(url=giturl, refspec="master")
    repo2.rev_parse() == repo.rev_parse()

    repo2.fetch_gerrit_notes(url=giturl)
    p = repo2.git_cmd(
        ["notes", "--ref=review", "show", new_sha], output=subprocess.PIPE
    )
    assert "invalid-note" in p.out



import logging
import os

from e3.anod.checkout import CheckoutManager
from e3.anod.status import ReturnValue
from e3.vcs.svn import SVNRepository
from e3.vcs.git import GitRepository
from e3.os.fs import touch


class TestCheckout:
    repo_data = os.path.join(os.path.dirname(__file__), "vcs_data")
    repo_data2 = os.path.join(os.path.dirname(__file__), "vcs_data2")

    def test_svn_checkout(self):
        os.environ["GIT_AUTHOR_EMAIL"] = "e3-core@example.net"
        os.environ["GIT_AUTHOR_NAME"] = "e3 core"
        os.environ["GIT_COMMITTER_NAME"] = "e3-core@example.net"
        os.environ["GIT_COMMITTER_EMAIL"] = "e3 core"

        url = SVNRepository.create("svn", initial_content_path=self.repo_data)
        url2 = SVNRepository.create("svn2", initial_content_path=self.repo_data2)
        url3 = GitRepository.create("git", initial_content_path=self.repo_data)
        url4 = GitRepository.create("git2", initial_content_path=self.repo_data2)

        r = SVNRepository(working_copy=os.path.abspath("svn_checkout"))
        r.update(url=url)

        m = CheckoutManager(name="myrepo", working_dir=".")

        logging.info("update of non existing url")
        result = m.update(vcs="svn", url=url + "wrong_url")
        assert result == ReturnValue.failure

        logging.info("update of existing url")
        result = m.update(vcs="svn", url=url)
        assert result == ReturnValue.success

        logging.info("update of existing url with no changes")
        result = m.update(vcs="svn", url=url)
        assert result == ReturnValue.unchanged

        # Update the repository
        logging.info("Do a checkin in svn repository")
        with open(os.path.join("svn_checkout", "file3.txt"), "w") as fd:
            fd.write("new file!")
        r.svn_cmd(["add", "file3.txt"])
        r.svn_cmd(["commit", "file3.txt", "-m", "checkin"])

        logging.info("Check that we see the update")
        result = m.update(vcs="svn", url=url)
        assert result == ReturnValue.success
        assert os.path.isfile(os.path.join("svn_checkout", "file3.txt"))

        logging.info("Do a local modification in the working dir")
        with open(os.path.join("myrepo", "file3.txt"), "w") as fd:
            fd.write("new file modified!")

        logging.info("And then do an update and check that cleanup was done")
        result = m.update(vcs="svn", url=url)
        assert result == ReturnValue.unchanged
        with open(os.path.join("myrepo", "file3.txt")) as fd:
            assert fd.read().strip() == "new file!"

        logging.info("Check that we can switch from one svn url to another")
        result = m.update(vcs="svn", url=url2)
        assert result == ReturnValue.success
        assert os.path.isfile(os.path.join("myrepo", "file1.txt", "data2.txt"))

        logging.info("Check that we can switch from one svn url to a git repo")
        result = m.update(vcs="git", url=url3, revision="master")
        assert result == ReturnValue.success
        assert os.path.isfile(os.path.join("myrepo", "file1.txt"))

        logging.info("Check that we can switch from one git url to another one")
        result = m.update(vcs="git", url=url4, revision="master")
        assert result == ReturnValue.success
        assert os.path.isfile(os.path.join("myrepo", "file1.txt", "data2.txt"))

        logging.info("Check that in case of no changes unchanged is returned")
        result = m.update(vcs="git", url=url4, revision="master")
        assert result == ReturnValue.unchanged

        logging.info("Check that changes are detected in git reposotories")
        with open(os.path.join("git2", "file3.txt"), "w") as fd:
            fd.write("new file!")
        r = GitRepository(os.path.abspath("git2"))
        r.git_cmd(["add", "file3.txt"])
        r.git_cmd(["commit", "-m", "new file"])
        result = m.update(vcs="git", url=url4, revision="master")
        assert result == ReturnValue.success
        assert os.path.isfile(os.path.join("myrepo", "file3.txt"))

        logging.info("Check that local modifications are discarded")
        with open(os.path.join("myrepo", "file3.txt"), "w") as fd:
            fd.write("new file modified!")
        result = m.update(vcs="git", url=url4, revision="master")
        assert result == ReturnValue.unchanged
        with open(os.path.join("myrepo", "file3.txt")) as fd:
            assert fd.read().strip() == "new file!"

        result = m.update(vcs="future-vcs", url="dummy")
        assert result == ReturnValue.failure

        result = m.update(vcs="git", url=url4 + "non-existing", revision="master")
        assert result == ReturnValue.failure

        result = m.update(vcs="git", url=url4)
        assert result == ReturnValue.failure

        # Add a .gitignore and use an external

        touch("git2/file4.txt")
        touch("git2/ignore_file.txt")
        with open("git2/.gitignore", "w") as fd:
            fd.write("/ignore_file.txt")

        result = m.update(vcs="external", url=os.path.abspath("git2"))
        assert os.path.isfile(os.path.join(m.working_dir, "file4.txt"))
        assert not os.path.isfile(os.path.join(m.working_dir, "ignore_file.txt"))
        assert result == ReturnValue.success

        result = m.update(vcs="external", url=os.path.abspath("git2"))
        assert result == ReturnValue.unchanged



from random import random

import e3.decorator

import pytest


def test_memoize():
    # First generate a function returning random values so that we
    # can see whether the cache is used or not

    @e3.decorator.memoize
    def t(arg):
        """Do foo."""
        del arg
        return random()

    assert "Do foo." in t.__repr__()

    assert t(1) == t(1)
    assert t(1) != t(2)

    # kwargs not supported
    with pytest.raises(TypeError):
        assert t(arg=1)

    # cache not used when non-hashable arguments
    assert t([1, 2, 3]) != t([1, 2, 3])

    # Verify that the cache is working also when
    # calling instance methods

    class C:
        @e3.decorator.memoize
        def t(self, arg):
            del arg
            return random()

    # same instance use the same cache
    c_instance = C()
    assert c_instance.t(2) == c_instance.t(2)

    # but not different instances
    assert C().t(2) != C().t(2)


def test_enabled():
    @e3.decorator.enabled
    def foo():
        return 22

    assert foo() == 22


def test_disabled():
    @e3.decorator.disabled
    def foo():
        return 22

    assert foo() is None



"""High-Level interface to SVN repository.

Example::
    svn = SVNRepository(working_copy='/tmp/working_dir')
    svn.update(url='file:///path_to_local_repo', revision=1,
               force_and_clean=True)
"""


from __future__ import annotations

import os.path
import re
import sys
from subprocess import PIPE
from typing import TYPE_CHECKING

import e3.log
import e3.os.fs
import e3.os.process
from e3.fs import mkdir, rm
from e3.vcs import VCSError

if TYPE_CHECKING:
    from typing import Any, IO, List, Optional, TextIO, Tuple, Union

    SVNCmd = List[Optional[str]]

logger = e3.log.getLogger("vcs.svn")


class SVNError(VCSError):
    pass


class SVNRepository:
    """Interface to a SVN Repository.

    :cvar svn_bin: path to the SVN binary
    :cvar log_stream: stream where the log commands will be redirected
        (default is stdout)
    :ivar working_copy: path to the SVN working copy
    """

    svn_bin = None
    log_stream: Union[TextIO, IO[str]] = sys.stdout

    def __init__(self, working_copy: str):
        """Initialize a SVNRepository object.

        :param working_copy: working copy of the SVNRepository
        """
        self.working_copy = working_copy

    @classmethod
    def is_unix_svn(cls) -> bool:
        """Check if svn is handling unix paths or windows paths.

        :return: True if unix paths should be used
        """
        if sys.platform != "win32":
            return True
        else:
            svn_version = e3.os.process.Run(["svn", "--version"]).out
            if "cygwin" in svn_version:
                return True
            else:
                return False

    @classmethod
    def local_url(cls, repo_path: str) -> str:
        """Return the url of a svn repository hosted locally.

        :param repo_path: path to the repo
        :return: the url that can be used as repository url
        """
        repo_path = os.path.abspath(repo_path)
        if not cls.is_unix_svn():
            if len(repo_path) > 1 and repo_path[1] == ":":
                # svn info returns the URL with an uppercase letter drive
                repo_path = repo_path[0].upper() + repo_path[1:]
            return "file:///" + repo_path.replace("\\", "/")
        else:  # windows: no cover
            return "file://" + e3.os.fs.unixpath(repo_path)

    @classmethod
    def create(cls, repo_path: str, initial_content_path: Optional[str] = None) -> str:
        """Create a local subversion repository.

        This creates a local repository (not a working copy) that can be
        referenced by using file:// protocol. The purpose of the this function
        is mainly to test svn-related functions without relying on a remote
        repository.

        :param repo_path: a local directory where to create the repository
        :param initial_content_path: directory containing the initial content
            of the repository. If set to None an empty repository is created.
        :return: the URL of the newly created repository
        """
        repo_path = os.path.abspath(repo_path)
        p = e3.os.process.Run(["svnadmin", "create", repo_path], output=cls.log_stream)

        if cls.is_unix_svn() and initial_content_path is not None:
            initial_content_path = e3.os.fs.unixpath(initial_content_path)

        if p.status != 0:
            raise SVNError(
                "cannot create svn repository in %s" % repo_path, origin="create"
            )
        if initial_content_path is not None:
            p = e3.os.process.Run(
                [
                    "svn",
                    "import",
                    initial_content_path,
                    cls.local_url(repo_path),
                    "-m",
                    "Initial import from %s" % initial_content_path,
                ],
                output=cls.log_stream,
            )
            if p.status != 0:
                raise SVNError(
                    "cannot perform initial import of %s into %s"
                    % (initial_content_path, repo_path),
                    origin="create",
                )
        return cls.local_url(repo_path)

    def svn_cmd(self, cmd: SVNCmd, **kwargs: Any) -> e3.os.process.Run:
        """Run a svn command.

        Add the non-interactive option to all command (accepted on all SVN.
        subcommands from version 1.5).
        :param cmd: the command line as a list of string, all None entries will
            be discarded
        :param kwargs: additional parameters to provide to e3.os.process.Run
        :return: Result of the Run of the SVN command
        :raise: SVNError
        """
        if self.__class__.svn_bin is None:
            svn_binary = e3.os.process.which("svn", default=None)
            if svn_binary is None:  # defensive code
                raise SVNError("cannot find svn", "svn_cmd")
            self.__class__.svn_bin = e3.os.fs.unixpath(svn_binary)

        if "output" not in kwargs:
            kwargs["output"] = self.log_stream

        p_cmd = [arg for arg in cmd if arg is not None]
        p_cmd.insert(0, self.__class__.svn_bin)
        p_cmd.append("--non-interactive")

        p = e3.os.process.Run(p_cmd, cwd=self.working_copy, **kwargs)
        if p.status != 0:
            raise SVNError(
                "{} failed (exit status: {})".format(
                    e3.os.process.command_line_image(p_cmd), p.status
                ),
                origin="svn_cmd",
                process=p,
            )
        return p

    def get_info(self, item: str) -> str:
        """Return a specific item shown by svn info.

        The --show-item option is only available from 1.9.
        :raise: SVNError
        """
        info = self.svn_cmd(["info"], output=PIPE).out
        m = re.search(fr"^{item}: *(.*)\n", info, flags=re.M)
        if m is None:
            logger.debug("svn info result:\n%s", info)
            raise SVNError(
                "Cannot fetch item %s from svn_info" % item, origin="get_info"
            )
        return m.group(1).strip()

    @property
    def url(self) -> str:
        """Return the last URL used for the checkout.

        :raise: SVNError
        """
        return self.get_info("URL")

    @property
    def current_revision(self) -> str:
        """Return the current revision.

        :raise: SVNError
        """
        try:
            return self.get_info("Last Changed Rev")
        except Exception:
            logger.exception("Cannot fetch last changed rev")
            raise SVNError("Cannot fetch last changed rev", "svn_cmd")

    def update(
        self,
        url: Optional[str] = None,
        revision: Optional[str] = None,
        force_and_clean: bool = False,
    ) -> bool:
        """Update a working copy or checkout a new one.

        If the directory is already a checkout, it tries to update it.
        If the directory is not associated to a (good) checkout or is empty
        it will checkout.
        The option --remove-unversioned of the svn subcommand
        cleanup exists only from svn version 1.9.
        :param url: URL of a SVN repository
        :param revision: specific revision (default is last)
        :param force_and_clean: if True: erase the content of non empty
        working_copy and use '--force' option for the svn update/checkout
        command
        :return: True if any local changes detected in the working copy
        :raise: SVNError
        """

        def is_clean_svn_dir(dir_path: str) -> Tuple[bool, bool]:
            """Return a tuple (True if dir is SVN directory, True if clean)."""
            if os.path.exists(os.path.join(dir_path, ".svn")):
                try:
                    status = self.svn_cmd(["status"], output=PIPE).out.strip()
                except SVNError:  # defensive code
                    return False, False
                if "warning: W" in status:
                    return False, False
                return True, status == ""
            return False, False

        def is_empty_dir(dir_path: str) -> bool:
            """Return True if the path is a directory and is empty."""
            return os.path.isdir(dir_path) and not os.listdir(dir_path)

        options: SVNCmd = ["--ignore-externals"]
        if revision:
            options += ["-r", revision]
        if force_and_clean:
            options += ["--force"]

        is_svn_dir, is_clean = is_clean_svn_dir(self.working_copy)
        if (
            is_svn_dir
            and (is_clean or not force_and_clean)
            and (not url or self.url == url)
        ):
            update_cmd: SVNCmd = ["update"]
            self.svn_cmd(update_cmd + options)
            return not is_clean
        if os.path.exists(self.working_copy):
            if not is_empty_dir(self.working_copy) and not force_and_clean:
                raise SVNError(
                    f"not empty {self.working_copy} url {url}", origin="update",
                )
            if is_svn_dir and not url:
                url = self.url
            rm(self.working_copy, recursive=True)

        mkdir(self.working_copy)
        checkout_cmd: SVNCmd = ["checkout", url, "."]
        self.svn_cmd(checkout_cmd + options)
        return not is_clean



from __future__ import annotations

import cgi
import contextlib
import json
import os
import socket
import tempfile
from collections import deque

import requests
import requests.adapters
import requests.exceptions
import requests.packages.urllib3.exceptions
from requests.packages.urllib3.util import Retry

from typing import TYPE_CHECKING

import e3.log
import requests_toolbelt.multipart
from e3.error import E3Error
from e3.fs import rm

if TYPE_CHECKING:
    from typing import Any, Callable, Deque, Dict, List, Optional, Tuple, Union
    from requests.auth import AuthBase
    from requests.models import Response


logger = e3.log.getLogger("net.http")


def get_filename(content_disposition: str) -> Optional[str]:
    """Return a filename from a HTTP Content-Disposition header.

    :param content_disposition: a Content-Disposition header string
    :return: the filename or None
    """
    _, value = cgi.parse_header(content_disposition)
    return value.get("filename")


class HTTPError(E3Error):
    pass


class BaseURL:
    """Represent a base url object along with its authentication.

    The root class BaseURL does not use authentication
    """

    def __init__(self, url: str):
        """Initialize a base url object.

        :param url: the base url
        """
        self.url = url

    def get_auth(self) -> Optional[Union[Tuple[str, str], AuthBase]]:
        """Return auth requests parameter.

        :return: authentication associated with the url
        """
        return None

    def __str__(self) -> str:
        return self.url


class HTTPSession:

    CHUNK_SIZE = 1024 * 1024
    DEFAULT_TIMEOUT = (60, 60)

    def __init__(self, base_urls: Optional[Union[List[str], List[BaseURL]]] = None):
        """Initialize HTTP session.

        :param base_urls: list of urls used as prefix to subsequent requests.
            Preferred base url is the first one in the list. In case of error
            during a request the next urls are used.
        """
        if base_urls:
            self.base_urls: Union[Deque[BaseURL], Deque[None]] = deque(
                [k if isinstance(k, BaseURL) else BaseURL(k) for k in base_urls]
            )
        else:
            self.base_urls = deque([None])
        self.session = requests.Session()
        self.last_base_url: Optional[BaseURL] = None

    def __enter__(self) -> HTTPSession:
        self.session.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.session.__exit__(exc_type, exc_val, exc_tb)

    def set_max_retries(
        self,
        base_url: Optional[str] = None,
        connect: Optional[int] = None,
        read: Optional[int] = None,
        redirect: Optional[int] = None,
    ) -> None:
        """Retry configuration.

        :param base_url: base url for the HTTPAdapter
        :param connect: how many connection-related errors to retry on
        :param read: how many times to retry on read errors
        :param redirect: how many redirects to perform. Limit this to avoid
            infinite redirect loops.
        """
        if base_url is None:
            base_urls = [str(b) for b in self.base_urls if b is not None]
        else:
            base_urls = [base_url]

        for url in base_urls:
            self.session.mount(
                url,
                requests.adapters.HTTPAdapter(
                    max_retries=Retry(connect=connect, read=read, redirect=redirect)
                ),
            )

    def request(
        self,
        method: str,
        url: str,
        data_streams: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Response:
        """Send a request.

        See requests Session.request function.

        The main difference is that several servers are tried in case
        base_urls have been set.

        For POST requests an additional parameter is supported: data_streams.
        data_streams is a dict associating a string key to either another
        string, a dict, a list or a file descriptor. String value are passed
        without any modifications. Lists and dicts are automatically encoded
        in JSON. Finally file objects are streamed during the POST request
        (no complete read is done into memory to fetch file content). When
        using data_streams parameter, data parameter will be ignored and
        headers one modified.
        """
        error_msgs = []

        for base_url in list(self.base_urls):
            if self.last_base_url != base_url:
                # if base_url is None then it means that self.base_urls is deque([None])
                # in that case last_base_url is always None. The code is confusing and
                # we should remove that complexity by always forcing self.base_urls to
                # be set. ???
                if TYPE_CHECKING:
                    assert base_url is not None
                self.session.auth = base_url.get_auth()
                self.last_base_url = base_url

            logger.debug("try with %s", base_url)

            # Handle data_streams. After some tests it seems that we need to
            # redo an instance of the multipart encoder for each attempt.
            if data_streams is not None:
                data: Dict[str, Any] = {}
                for k, v in data_streams.items():
                    if hasattr(v, "seek"):
                        # This is a file. Assume that the key is the filename
                        data[k] = (k, v)
                        v.seek(0)
                    elif isinstance(v, (dict, list)):
                        # Automatically encode to json
                        data[k] = json.dumps(v)
                    else:
                        data[k] = v
                kwargs["data"] = requests_toolbelt.multipart.encoder.MultipartEncoder(
                    data
                )
                header = {"Content-Type": kwargs["data"].content_type}
                if "headers" in kwargs:
                    kwargs["headers"].update(header)
                else:
                    kwargs["headers"] = header

            # Compute final url
            if base_url is not None:
                final_url = f"{base_url}/{url}"
                message_prefix = "%s: " % base_url
            else:
                final_url = url
                message_prefix = ""

            if "timeout" not in kwargs:
                kwargs["timeout"] = self.DEFAULT_TIMEOUT

            try:
                logger.debug("%s %s", method, final_url)
                response = self.session.request(method, final_url, **kwargs)
                if response.status_code != 200:
                    error_msgs.append(f"{message_prefix}{response.text}")
                    response.raise_for_status()
                return response
            except (
                socket.timeout,
                requests.exceptions.RequestException,
                requests.packages.urllib3.exceptions.HTTPError,
            ) as e:
                # got an error with that base url so put it last in our list
                error_msgs.append(f"{message_prefix}{e}")
                problematic_url = self.base_urls.popleft()
                self.base_urls.append(problematic_url)  # type: ignore

        raise HTTPError(
            "got request error (%d):\n%s" % (len(error_msgs), "\n".join(error_msgs))
        )

    def download_file(
        self,
        url: str,
        dest: str,
        filename: Optional[str] = None,
        validate: Optional[Callable[[str], bool]] = None,
    ) -> Optional[str]:
        """Download a file.

        :param url: the url to GET
        :param dest: local directory path for the downloaded file
        :param filename: the local path whether to store this resource, by
            default use the name provided  in the ``Content-Disposition``
            header.
        :param validate: function to call once the download is complete for
            detecting invalid / corrupted download. Takes the local path as
            parameter and returns a boolean.
        :return: the name of the file or None if there is an error
        """
        # When using stream=True, Requests cannot release the connection back
        # to the pool unless all the data is consumed or Response.close called.
        # Force Response.close by wrapping the code with contextlib.closing

        path = None
        try:
            with contextlib.closing(
                self.request(method="GET", url=url, stream=True)
            ) as response:
                content_length = int(response.headers.get("content-length", 0))
                e3.log.debug(response.headers)
                if filename is None:
                    if "content-disposition" in response.headers:
                        filename = get_filename(response.headers["content-disposition"])
                    if filename is None:
                        # Generate a temporary name
                        tmpf = tempfile.NamedTemporaryFile(
                            delete=False, dir=dest, prefix="download."
                        )
                        tmpf.close()
                        filename = tmpf.name

                path = os.path.join(dest, filename)
                logger.info("downloading %s size=%s", path, content_length)

                expected_size = content_length // self.CHUNK_SIZE
                with open(path, "wb") as fd:
                    for chunk in e3.log.progress_bar(
                        response.iter_content(self.CHUNK_SIZE), total=expected_size
                    ):
                        fd.write(chunk)
                if validate is None or validate(path):
                    return path
                else:
                    rm(path)
        except (requests.exceptions.RequestException, HTTPError) as e:
            # An error (timeout?) occurred while downloading the file
            logger.warning("download failed")
            logger.debug(e)
            if path is not None:
                rm(path)
        return None
